# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM 3 conversion extras.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.passes.qasm.compat import add_stdgates_include, insert_gate_def, replace_gate_names
from qbraid.transpiler.annotations import requires_extras

qbraid_qir = LazyLoader("qbraid_qir", globals(), "qbraid_qir")
autoqasm = LazyLoader("autoqasm", globals(), "autoqasm")

if TYPE_CHECKING:
    import autoqasm as aq

aq_to_qasm3_stdgates = {
    "ccnot": "ccx",
    "cnot": "cx",
    "cphaseshift": "cp",
    "i": "id",
    "phaseshift": "p",
    "si": "sdg",
    "ti": "tdg",
    "v": "sx",
    "vi": "sxdg",
}


if TYPE_CHECKING:
    import pyqir

    from qbraid.programs.typer import Qasm3StringType


@requires_extras("qbraid_qir")
def qasm3_to_pyqir(program: Qasm3StringType) -> pyqir.Module:
    """Returns a PyQIR module equivalent to the input OpenQASM 3 program.

    Args:
        program (str): OpenQASM 3 program circuit to convert to PyQIR module.

    Returns:
        pyqir.Module: module equivalent to input OpenQASM 3 program.
    """
    return qbraid_qir.qasm3.qasm3_to_qir(program)


@requires_extras("autoqasm")
def autoqasm_to_qasm3(program: aq.program.program.Program) -> Qasm3StringType:
    """Converts an AutoQASM program to an OpenQASM 3 program of gates from the
    Standard Gate Library. The program must be built prior to conversion using
    the .build() method.

    Args:
        program (autoqasm.program.program.Program): AutoQASM program to convert
            to OpenQASM 3.

    Returns:
        str: OpenQASM 3 program equivalent to input AutoQASM program.
    """
    qasm = program.to_ir()
    # Convert to Standard Library qasm3 gates
    qasm = replace_gate_names(qasm, aq_to_qasm3_stdgates)
    # Insert custom gate conversions
    qasm = insert_gate_def(qasm, ["iswap", "sxdg", "cv"])
    # AutoQASM does not include stdgates.inc
    qasm = add_stdgates_include(qasm)
    return qasm



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module used for loading entry points and other dynamic imports.

"""
import importlib.metadata
from typing import Optional, Type

from .exceptions import QbraidError


def get_entrypoints(module: str) -> dict[str, object]:
    """
    Retrieves entry points for a given module.

    Args:
        module (str): The name of the module to retrieve entry points for.

    Returns:
        dict[str, object]: Dictionary of entry point names and their associated entry point objects.
    """
    group = f"qbraid.{module}"

    entry_points = importlib.metadata.entry_points().select(group=group)

    return {ep.name: ep for ep in entry_points}


def load_entrypoint(module: str, name: str) -> Optional[Type]:
    """
    Load an entrypoint given its module and name.

    Args:
        module (str): Module of entrypoint to load, e.g., "programs"
        name (str): Name of the entrypoint to load within the module.

    Returns:
        Optional[Type]: The loaded entry point class, or None if not found or failed to load.

    Raises:
        ValueError: If the specified entry point cannot be found.
        QbraidError: If the specified entry point fails to load.
    """
    try:
        entry_points = get_entrypoints(module)
        entry_point = entry_points[name]
        return entry_point.load()
    except KeyError as err:
        raise ValueError(f"Entrypoint '{name}' not found in module '{module}'.") from err
    except Exception as err:
        raise QbraidError(f"Failed to load entrypoint '{name}' from module '{module}'.") from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for lazy loading of modules, objects and entry points

"""
from unittest.mock import MagicMock, patch

import pytest

import qbraid
from qbraid._entrypoints import get_entrypoints, load_entrypoint
from qbraid.exceptions import QbraidError
from qbraid.programs._import import _dynamic_importer


def test_load_entrypoint_success():
    """Test that an entrypoint is successfully loaded."""
    module = "programs"
    name = "qasm3"
    assert load_entrypoint(module, name).__name__ == "OpenQasm3Program"


def test_load_entrypoint_raise_value_error():
    """Test that a ValueError is raised when the entrypoint is not found."""
    module = "programs"
    name = "nonexistent"

    with pytest.raises(ValueError) as excinfo:
        load_entrypoint(module, name)
    assert f"Entrypoint '{name}' not found in module '{module}'." in str(excinfo.value)


def test_load_entrypoint_raises_qbraiderror_on_general_exception():
    """Test that a QbraidError is raised when loading the entrypoint fails."""
    mock_entry_point = MagicMock()
    mock_entry_point.load.side_effect = Exception("Some error")
    with patch("qbraid._entrypoints.get_entrypoints", return_value={"test": mock_entry_point}):
        with pytest.raises(
            QbraidError, match="Failed to load entrypoint 'test' from module 'programs'"
        ):
            load_entrypoint("programs", "test")


def mock_get_class(name):
    return type(name, (), {})


def mock_assign_default_type_alias(imported, program_type):
    return "mock_alias"


def test_dynamic_importer_exception():
    opt_modules = ["nonexistent.module"]

    with patch("qbraid.programs._import.import_module", side_effect=Exception("Import error")):
        with patch("qbraid.programs._import._get_class", side_effect=mock_get_class):
            with patch(
                "qbraid.programs._import._assign_default_type_alias",
                side_effect=mock_assign_default_type_alias,
            ):
                imported = _dynamic_importer(opt_modules)
                assert imported == {}


@patch("qbraid._entrypoints.importlib.metadata.entry_points")
def test_get_entrypoints_with_new_python_version(mock_importlib_eps):
    """Test that entry points are retrieved correctly with Python 3.10+."""
    mock_entry_point = MagicMock()
    mock_entry_point.name = "qasm2"
    mock_importlib_eps.return_value.select.return_value = [mock_entry_point]

    result = get_entrypoints("programs")

    assert result == {"qasm2": mock_entry_point}

    mock_importlib_eps.return_value.select.assert_called_once_with(group="qbraid.programs")


@pytest.mark.parametrize("module_name", list(qbraid._lazy.keys()))
def test_lazy_loading_modules(module_name):
    """Test lazy loading of modules."""
    mod = getattr(qbraid, module_name)
    assert mod is not None
    assert mod.__name__.endswith(f".{module_name}")


@pytest.mark.parametrize(
    "obj_name", [item for sublist in qbraid._lazy.values() for item in sublist]
)
def test_lazy_loading_objects(obj_name):
    """Test lazy loading of objects."""
    obj = getattr(qbraid, obj_name)
    assert obj is not None
    assert hasattr(qbraid, obj_name)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing interface for submitting and managing
quantum jobs through (native) qBraid APIs.

.. currentmodule:: qbraid.runtime.native

Classes
--------

.. autosummary::
   :toctree: ../stubs/

    Session
    QbraidSession
    QbraidClient
    QbraidProvider
    QbraidDevice
    QbraidJob
    QirRunner

ResultData Subclasses
^^^^^^^^^^^^^^^^^^^^^^^

.. autosummary::
   :toctree: ../stubs/

    QuEraQasmSimulatorResultData
    QbraidQirSimulatorResultData
    NECVectorAnnealerResultData
    Equal1SimulatorResultData

"""
from qbraid_core import QbraidClient, QbraidSession, Session
from qbraid_core.services.quantum.runner import QirRunner

from .device import QbraidDevice
from .job import QbraidJob
from .provider import QbraidProvider
from .result import (
    Equal1SimulatorResultData,
    NECVectorAnnealerResultData,
    QbraidQirSimulatorResultData,
    QuEraQasmSimulatorResultData,
)

__all__ = [
    "Session",
    "QbraidSession",
    "QbraidClient",
    "QbraidProvider",
    "QbraidDevice",
    "QbraidJob",
    "QirRunner",
    "QuEraQasmSimulatorResultData",
    "QbraidQirSimulatorResultData",
    "NECVectorAnnealerResultData",
    "Equal1SimulatorResultData",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QiskitGateModelResultBuilder Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import numpy as np
from qiskit.exceptions import QiskitError
from qiskit.primitives.containers import PrimitiveResult

from qbraid._logging import logger
from qbraid.runtime.postprocess import normalize_tuples

if TYPE_CHECKING:
    from qiskit.result import Result
    from qiskit_ibm_runtime.utils.runner_result import RunnerResult


class QiskitGateModelResultBuilder:
    """Qiskit ``Result`` wrapper class."""

    def __init__(self, result: PrimitiveResult | RunnerResult | Result):
        self._result = result

    def _format_measurements(self, memory_list: list[str]) -> list[list[int]]:
        """Format the measurements into int for the given memory list"""
        formatted_meas = []
        for str_shot in memory_list:
            lst_shot = [int(x) for x in list(str_shot)]
            formatted_meas.append(lst_shot)
        return formatted_meas

    def measurements(self) -> Optional[np.ndarray | list[np.ndarray]]:
        """Return measurements a 2D numpy array"""
        if isinstance(self._result, PrimitiveResult):
            meas = [pub_result.join_data().array for pub_result in self._result]
            return meas[0] if len(meas) == 1 else meas

        num_circuits = len(self._result.results)

        try:
            qiskit_meas = [self._result.get_memory(i) for i in range(num_circuits)]
        except QiskitError as err:
            logger.warning("Memory states (measurements) data not available for this job: %s", err)
            return None

        qbraid_meas = [self._format_measurements(qiskit_meas[i]) for i in range(num_circuits)]

        if num_circuits == 1:
            qbraid_meas = qbraid_meas[0]
        else:
            qbraid_meas = normalize_tuples(qbraid_meas)

        return np.array(qbraid_meas)

    def get_counts(self) -> dict[str, int] | list[dict[str, int]]:
        """Returns the histogram data of the run"""
        if isinstance(self._result, PrimitiveResult):
            counts = [pub_result.join_data().get_counts() for pub_result in self._result]
            return counts[0] if len(counts) == 1 else counts
        return self._result.get_counts()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Qiskit conversion extras.

"""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import requires_extras

qiskit_braket_provider = LazyLoader("qiskit_braket_provider", globals(), "qiskit_braket_provider")
qiskit_qir = LazyLoader("qiskit_qir", globals(), "qiskit_qir")
qiskit_ionq = LazyLoader("qiskit_ionq", globals(), "qiskit_ionq")

if TYPE_CHECKING:
    import braket.circuits
    import pyqir
    import qiskit.circuit

    import qbraid.programs


@requires_extras("qiskit_braket_provider")
def qiskit_to_braket(circuit: qiskit.circuit.QuantumCircuit, **kwargs) -> braket.circuits.Circuit:
    """Return a Braket quantum circuit from a Qiskit quantum circuit.

    Args:
        circuit (QuantumCircuit): Qiskit quantum circuit
        basis_gates (Optional[Iterable[str]]): The gateset to transpile to.
            If `None`, the transpiler will use all gates defined in the Braket SDK.
            Default: `None`.
        verbatim (bool): Whether to translate the circuit without any modification, in other
            words without transpiling it. Default: False.

    Returns:
        Circuit: Braket circuit
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        return qiskit_braket_provider.providers.adapter.to_braket(circuit, **kwargs)


@requires_extras("qiskit_qir")
def qiskit_to_pyqir(circuit: qiskit.circuit.QuantumCircuit) -> pyqir.Module:
    """Return a PyQIR module from a Qiskit quantum circuit.

    Args:
        circuit (QuantumCircuit): Qiskit quantum circuit

    Returns:
        Module: PyQIR module
    """
    # tuple of module and list of entry points
    module, _ = qiskit_qir.to_qir_module(circuit)
    return module


@requires_extras("qiskit_ionq")
def qiskit_to_ionq(circuit: qiskit.circuit.QuantumCircuit, **kwargs) -> qbraid.programs.IonQDict:
    """Return a IonQDict from a Qiskit quantum circuit.

    Args:
        circuit (QuantumCircuit): Qiskit quantum circuit

    Returns:
        dict: IonQDict representing the circuit
    """
    # pylint: disable-next=import-outside-toplevel
    from qbraid.programs.gate_model.ionq import GateSet, InputFormat

    instrs, _, _ = qiskit_ionq.helpers.qiskit_circ_to_ionq_circ(circuit, **kwargs)
    return {
        "format": InputFormat.CIRCUIT.value,
        "gateset": kwargs.get("gateset", GateSet.QIS.value),
        "qubits": circuit.num_qubits,
        "circuit": instrs,
    }



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for the about module.

"""

import sys
from importlib.metadata import PackageNotFoundError, distribution, version
from io import StringIO
from unittest.mock import patch

import numpy
import openqasm3
import pydantic
import pyqasm
import qbraid_core
import rustworkx

from qbraid import _about
from qbraid._version import __version__

PYQASM_VERSION = version("pyqasm")
PYQASM_VERSION_BUG = PYQASM_VERSION in {"0.2.0", "0.2.1"}

TYPING_EXTS_VERSION = version("typing-extensions")


def test_about():
    """Test the about function."""

    # Redirect stdout to capture the output of the about function
    captured_output = StringIO()
    sys.stdout = captured_output

    _about.about()

    # Restore stdout
    sys.stdout = sys.__stdout__

    # Get the captured output as a string
    output = captured_output.getvalue()

    core_dependencies = {
        "numpy": numpy.__version__,
        "openqasm3": openqasm3.__version__,
        "pyqasm": pyqasm.__version__ if not PYQASM_VERSION_BUG else PYQASM_VERSION,
        "pydantic": pydantic.__version__,
        "qbraid-core": qbraid_core.__version__,
        "rustworkx": rustworkx.__version__,
        "typing-extensions": TYPING_EXTS_VERSION,
    }

    assert f"qbraid:\t{__version__}\n\n" in output
    assert "Core Dependencies" in output
    for dep, v in core_dependencies.items():
        assert f"{dep}: {v}" in output
    assert "Optional Dependencies" in output
    assert "Python:" in output
    assert "Platform:" in output


def test_about_no_optional_dependencies():
    """Test the about function when no optional dependencies are available."""
    _, extras = _about.get_dependencies("qbraid")
    optional_packages = {item for subset in extras.values() for item in subset}

    def custom_distribution(name):
        if any(dep in name for dep in optional_packages):
            raise PackageNotFoundError
        return distribution(name)

    with patch("importlib.metadata.distribution", side_effect=custom_distribution):
        with patch("builtins.print") as mock_print:
            _about.about()
            actual_print_args = " ".join([str(arg[0]) for arg in mock_print.call_args_list])
            assert "None" in actual_print_args



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for providing transforamtions with basis gates.
across various other quantum software frameworks.

"""
from typing import Optional, Union

from openqasm3 import ast, dumps
from openqasm3.parser import QASM3ParsingError, parse

from qbraid.passes.exceptions import CompilationError, QasmDecompositionError

from .compat import _replace_gate_names, declarations_to_qasm2


def _decompose_crx(gate: ast.QuantumGate) -> list[ast.Statement]:
    """Decompose a crx gate into its basic gate equivalents."""
    theta = gate.arguments[0]
    control = gate.qubits[0]
    target = gate.qubits[1]

    rz_pos_pi_half = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="rz"),
        arguments=[
            ast.BinaryExpression(
                op=ast.BinaryOperator(17),
                lhs=ast.Identifier(name="pi"),
                rhs=ast.FloatLiteral(value=2),
            )
        ],
        qubits=[target],
    )
    ry_pos_theta_half = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="ry"),
        arguments=[
            ast.BinaryExpression(
                op=ast.BinaryOperator(17), lhs=theta, rhs=ast.FloatLiteral(value=2)
            )
        ],
        qubits=[target],
    )
    cx = ast.QuantumGate(
        modifiers=[], name=ast.Identifier(name="cx"), arguments=[], qubits=[control, target]
    )
    ry_neg_theta_half = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="ry"),
        arguments=[
            ast.BinaryExpression(
                op=ast.BinaryOperator(17),
                lhs=ast.UnaryExpression(ast.UnaryOperator(3), theta),
                rhs=ast.FloatLiteral(value=2),
            )
        ],
        qubits=[target],
    )
    rz_neg_pi_half = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="rz"),
        arguments=[
            ast.BinaryExpression(
                op=ast.BinaryOperator(17),
                lhs=ast.UnaryExpression(ast.UnaryOperator(3), ast.Identifier(name="pi")),
                rhs=ast.FloatLiteral(value=2),
            )
        ],
        qubits=[target],
    )
    return [rz_pos_pi_half, ry_pos_theta_half, cx, ry_neg_theta_half, cx, rz_neg_pi_half]


def _decompose_cry(gate: ast.QuantumGate) -> list[ast.Statement]:
    """Decompose a cry gate into its basic gate equivalents."""
    theta = gate.arguments[0]
    control = gate.qubits[0]
    target = gate.qubits[1]

    ry_pos_theta_half = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="ry"),
        arguments=[
            ast.BinaryExpression(
                op=ast.BinaryOperator(17), lhs=theta, rhs=ast.FloatLiteral(value=2)
            )
        ],
        qubits=[target],
    )
    cx = ast.QuantumGate(
        modifiers=[], name=ast.Identifier(name="cx"), arguments=[], qubits=[control, target]
    )
    ry_neg_theta_half = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="ry"),
        arguments=[
            ast.UnaryExpression(
                ast.UnaryOperator(3),
                ast.BinaryExpression(
                    op=ast.BinaryOperator(17), lhs=theta, rhs=ast.FloatLiteral(value=2)
                ),
            )
        ],
        qubits=[target],
    )
    return [ry_pos_theta_half, cx, ry_neg_theta_half, cx]


def _decompose_crz(gate: ast.QuantumGate) -> list[ast.Statement]:
    """Decompose a cry gate into its basic gate equivalents."""
    theta = gate.arguments[0]
    control = gate.qubits[0]
    target = gate.qubits[1]

    rz_pos_theta_half = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="rz"),
        arguments=[
            ast.BinaryExpression(
                op=ast.BinaryOperator(17), lhs=theta, rhs=ast.FloatLiteral(value=2)
            )
        ],
        qubits=[target],
    )
    cx = ast.QuantumGate(
        modifiers=[], name=ast.Identifier(name="cx"), arguments=[], qubits=[control, target]
    )
    rz_neg_theta_half = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="rz"),
        arguments=[
            ast.UnaryExpression(
                ast.UnaryOperator(3),
                ast.BinaryExpression(
                    op=ast.BinaryOperator(17), lhs=theta, rhs=ast.FloatLiteral(value=2)
                ),
            )
        ],
        qubits=[target],
    )
    return [rz_pos_theta_half, cx, rz_neg_theta_half, cx]


def _decompose_cy(gate: ast.QuantumGate, *args) -> list[ast.Statement]:
    """Decompose a cy gate into its basic gate equivalents."""
    control = gate.qubits[0]
    target = gate.qubits[1]

    cry_pi = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="cry"),
        arguments=[ast.Identifier(name="pi")],
        qubits=[control, target],
    )
    s = ast.QuantumGate(modifiers=[], name=ast.Identifier(name="s"), arguments=[], qubits=[control])
    return decompose(ast.Program(statements=[cry_pi, s]), *args).statements


def _decompose_cz(gate: ast.QuantumGate) -> list[ast.Statement]:
    """Decompose a cz gate into its basic gate equivalents."""
    control = gate.qubits[0]
    target = gate.qubits[1]

    crz_pi = ast.QuantumGate(
        modifiers=[],
        name=ast.Identifier(name="crz"),
        arguments=[ast.Identifier(name="pi")],
        qubits=[control, target],
    )
    s = ast.QuantumGate(modifiers=[], name=ast.Identifier(name="s"), arguments=[], qubits=[control])
    return decompose(ast.Program(statements=[crz_pi, s])).statements


def decompose(program: ast.Program, gateset: Optional[set[str]] = None) -> ast.Program:
    """Decompose a program into its basic gate equivalents."""
    decomposition_map = {
        "crx": _decompose_crx,
        "cry": _decompose_cry,
        "crz": _decompose_crz,
        "cy": _decompose_cy,
        "cz": _decompose_cz,
    }

    transformed_statements = []
    for statement in program.statements:
        if isinstance(statement, ast.QuantumGate):
            gate_name = statement.name.name
            if gate_name in decomposition_map and (gateset is None or gate_name not in gateset):
                transformed_statements.extend(decomposition_map[gate_name](statement))
            else:
                transformed_statements.append(statement)
        else:
            transformed_statements.append(statement)

    return ast.Program(statements=transformed_statements, version=program.version)


def assert_gates_in_basis(program: ast.Program, gateset: set[str]) -> None:
    """Verify that the program is represented only by gates in the given basis gate set."""
    for statement in program.statements:
        if isinstance(statement, ast.QuantumGate):
            gate_name = statement.name.name
            if gate_name not in gateset:
                raise ValueError(
                    f"OpenQASM program uses gate '{gate_name}' which is not in the basis gate set."
                )


def rebase(
    qasm: str,
    gateset: Union[set[str], str],
    require_predicates: bool = True,
    gate_mappings: Optional[dict[str, str]] = None,
    case_sensitive: bool = False,
) -> str:
    """
    Rebases an OpenQASM 3 program according to a given basis gate set.

    Args:
        qasm (str): The original OpenQASM 3 program as a string.
        gateset (set[str]): The target basis gates to decompose the program to.
        require_predicates (bool): If True, raises an error if the program fails to meet compilation
            predicates. If False, returns the original program on failure. Defaults to True.
        gate_mappings (dict[str, str]): A dictionary mapping gate names to new gate names.
        case_sensitive (bool): If True, the gate mappings are case-sensitive. Defaults to False.

    Returns:
        str: The decomposed OpenQASM 3 program.

    Raises:
        ValueError: If no basis gates are provided or if the basis gate set identifier is invalid
        TypeError: If the basis gate set is not a set of strings or a string identifier
        QasmDecompositionError: If an error occurrs during the decomposition process
        CompilationError: If the program cannot be rebased to the provided basis gate set

    """
    # Validate basis gates
    if isinstance(gateset, set):
        if len(gateset) == 0:
            raise ValueError("Basis gate set cannot be empty.")
    elif isinstance(gateset, str):
        if gateset.lower() == "any":
            gateset = set()
        else:
            raise ValueError("Invalid basis gate set identifier.")
    else:
        raise TypeError("Basis gate set must be a set of strings or a string identifier.")

    # Parse program and apply decomposition(s)
    try:
        program = parse(qasm)
    except QASM3ParsingError as err:
        raise ValueError("Invalid OpenQASM program.") from err

    try:
        converted_program = decompose(program, gateset)
    except Exception as err:  # pylint: disable=broad-exception-caught
        raise QasmDecompositionError from err

    if gate_mappings is not None:
        converted_program = _replace_gate_names(converted_program, gate_mappings, case_sensitive)
        if require_predicates:
            gateset = {gate_mappings.get(gate, gate) for gate in gateset}

    # Check if the program meets the compilation predicates
    try:
        if len(gateset) > 0:
            assert_gates_in_basis(converted_program, gateset)
    except ValueError as err:
        if require_predicates:
            raise CompilationError(
                "Rebasing the specified quantum program to the provided "
                f"basis gate set {gateset} is not supported."
            ) from err

        return qasm

    version_major = converted_program.version.split(".")[0]
    qasm = dumps(converted_program)

    if int(version_major) == 2:
        qasm = declarations_to_qasm2(qasm)

    return qasm


__all__ = ["decompose", "rebase", "assert_gates_in_basis"]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing PyTKET conversion extras.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import requires_extras

pytket_braket = LazyLoader("pytket_braket", globals(), "pytket.extensions.braket")

if TYPE_CHECKING:
    import braket.circuits
    import pytket.circuit


@requires_extras("pytket.extensions.braket")
def pytket_to_braket(circuit: pytket.circuit.Circuit) -> braket.circuits.Circuit:
    """Returns an Amazon Braket circuit equivalent to the input pytket circuit.

    Args:
        circuit (pytket.circuit.Circuit): PyTKET circuit to convert to Braket circuit.

    Returns:
        braket.circuits.Circuit: Braket circuit equivalent to input pytket circuit.
    """
    braket_circuit, _, _ = pytket_braket.braket_convert.tk_to_braket(circuit)
    return braket_circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing qasm programs used for testing

"""
import os

import openqasm3

QASMType = str
current_dir = os.path.dirname(os.path.abspath(__file__))


def _read_qasm_file(filename: str) -> QASMType:
    """Reads a qasm file from the qasm_lib directory"""
    return open(os.path.join(current_dir, filename), mode="r", encoding="utf-8").read()


def qasm3_bell() -> QASMType:
    """Returns OpenQASM3 bell circuit"""
    return _read_qasm_file("bell.qasm")


def qasm3_shared15() -> QASMType:
    """Returns OpenQASM3 15 gate test circuit."""
    return _read_qasm_file("shared15.qasm")


def openqasm3_bell():
    """Returns OpenQASM3 AST representing bell program"""
    qasm_str = qasm3_bell()
    return openqasm3.parse(qasm_str)


def openqasm3_shared15():
    """Returns OpenQASM3 AST representing 15 gate test program"""
    qasm_str = qasm3_shared15()
    return openqasm3.parse(qasm_str)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Benchmarking tests for PyTKET conversions

"""
import importlib.util
import string

import numpy as np
import pytest
from cirq.contrib.qasm_import import circuit_from_qasm

from qbraid.interface import circuits_allclose
from qbraid.transpiler import ConversionGraph, transpile

try:
    import pytket

    pytket_not_installed = False
except ImportError:
    pytket_not_installed = True

pytestmark = pytest.mark.skipif(pytket_not_installed, reason="pytket not installed")

np.random.seed(0)

# TODO: Investigate generating params dynamically.
# Difficult because pytket methods are wrapped around C++ code.
gates_param_map = {
    "H": {"qubit": 0},
    "S": {"qubit": 0},
    "SX": {"qubit": 0},
    "SXdg": {"qubit": 0},
    "Sdg": {"qubit": 0},
    "T": {"qubit": 0},
    "Tdg": {"qubit": 0},
    "V": {"qubit": 0},
    "Vdg": {"qubit": 0},
    "X": {"qubit": 0},
    "Y": {"qubit": 0},
    "Z": {"qubit": 0},
    "Phase": {"arg0": np.random.rand() * 2 * np.pi},
    "Rx": {"angle": np.random.rand() * 2 * np.pi, "qubit": 0},
    "Ry": {"angle": np.random.rand() * 2 * np.pi, "qubit": 0},
    "Rz": {"angle": np.random.rand() * 2 * np.pi, "qubit": 0},
    "U1": {"angle": np.random.rand() * 2 * np.pi, "qubit": 0},
    "ECR": {"qubit_0": 0, "qubit_1": 1},
    "SWAP": {"qubit_0": 0, "qubit_1": 1},
    "ISWAPMax": {"qubit0": 0, "qubit1": 1},
    "Sycamore": {"qubit0": 0, "qubit1": 1},
    "ZZMax": {"qubit0": 0, "qubit1": 1},
    "CH": {"control_qubit": 0, "target_qubit": 1},
    "CSX": {"control_qubit": 0, "target_qubit": 1},
    "CSXdg": {"control_qubit": 0, "target_qubit": 1},
    "CV": {"control_qubit": 0, "target_qubit": 1},
    "CVdg": {"control_qubit": 0, "target_qubit": 1},
    "CX": {"control_qubit": 0, "target_qubit": 1},
    "CY": {"control_qubit": 0, "target_qubit": 1},
    "CZ": {"control_qubit": 0, "target_qubit": 1},
    "CCX": {"control_0": 0, "control_1": 1, "target": 2},
    "CSWAP": {"control": 0, "target_0": 1, "target_1": 2},
    "Measure": {"qubit": 0, "bit_index": 0},
    "CRx": {"angle": np.random.rand() * 2 * np.pi, "control_qubit": 0, "target_qubit": 1},
    "CRy": {"angle": np.random.rand() * 2 * np.pi, "control_qubit": 0, "target_qubit": 1},
    "CRz": {"angle": np.random.rand() * 2 * np.pi, "control_qubit": 0, "target_qubit": 1},
    "CU1": {"angle": np.random.rand() * 2 * np.pi, "control_qubit": 0, "target_qubit": 1},
    "CU3": {
        "angle0": np.random.rand() * 2 * np.pi,
        "angle1": np.random.rand() * 2 * np.pi,
        "angle2": np.random.rand() * 2 * np.pi,
        "control_qubit": 0,
        "target_qubit": 1,
    },
    "ESWAP": {"angle": np.random.rand() * 2 * np.pi, "qubit0": 0, "qubit1": 1},
    "ISWAP": {"angle": np.random.rand() * 2 * np.pi, "qubit0": 0, "qubit1": 1},
    "FSim": {
        "angle0": np.random.rand() * 2 * np.pi,
        "angle1": np.random.rand() * 2 * np.pi,
        "qubit0": 0,
        "qubit1": 1,
    },
    "PhasedISWAP": {
        "angle0": np.random.rand() * 2 * np.pi,
        "angle1": np.random.rand() * 2 * np.pi,
        "qubit0": 0,
        "qubit1": 1,
    },
    "PhasedX": {
        "angle0": np.random.rand() * 2 * np.pi,
        "angle1": np.random.rand() * 2 * np.pi,
        "qubit": 0,
    },
    "U2": {
        "angle0": np.random.rand() * 2 * np.pi,
        "angle1": np.random.rand() * 2 * np.pi,
        "qubit": 0,
    },
    "TK1": {
        "angle0": np.random.rand() * 2 * np.pi,
        "angle1": np.random.rand() * 2 * np.pi,
        "angle2": np.random.rand() * 2 * np.pi,
        "qubit": 0,
    },
    "U3": {
        "angle0": np.random.rand() * 2 * np.pi,
        "angle1": np.random.rand() * 2 * np.pi,
        "angle2": np.random.rand() * 2 * np.pi,
        "qubit": 0,
    },
    "TK2": {
        "angle0": np.random.rand() * 2 * np.pi,
        "angle1": np.random.rand() * 2 * np.pi,
        "angle2": np.random.rand() * 2 * np.pi,
        "qubit0": 0,
        "qubit1": 1,
    },
    "XXPhase": {"angle": np.random.rand() * 2 * np.pi, "qubit0": 0, "qubit1": 1},
    "YYPhase": {"angle": np.random.rand() * 2 * np.pi, "qubit0": 0, "qubit1": 1},
    "ZZPhase": {"angle": np.random.rand() * 2 * np.pi, "qubit0": 0, "qubit1": 1},
    "XXPhase3": {"angle": np.random.rand() * 2 * np.pi, "qubit0": 0, "qubit1": 1, "qubit2": 2},
}


@pytest.fixture
def pytket_circuits():
    """Construct a dictionary of PyTKET circuits over all supported gates
    using random parameters as applicable.
    """
    pytket_gates = {attr: None for attr in dir(pytket.Circuit) if attr[0] in string.ascii_uppercase}
    for gate in pytket_gates:
        try:
            if gates_param_map[gate] is None:
                continue
            pytket_gates[gate] = getattr(pytket.Circuit(3, 1), gate)(**gates_param_map[gate])
        except Exception:  # pylint: disable=broad-exception-caught
            continue
    return {k: v for k, v in pytket_gates.items() if v is not None}


@pytest.fixture
def conversion_graph():
    """Return a conversion graph of natively supported conversions."""
    return ConversionGraph(require_native=True)


def is_package_installed(package_name: str) -> bool:
    """Check if a package is installed."""
    return importlib.util.find_spec(package_name) is not None


ALL_TARGETS = [("braket", 0.64), ("cirq", 0.66), ("pyquil", 0.66), ("qiskit", 0.66)]
AVAILABLE_TARGETS = [(name, version) for name, version in ALL_TARGETS if is_package_installed(name)]


def convert_from_pytket_to_x(target, circuit_name, circuits, graph):
    """Construct a PyTKET circuit with the given gate, transpile it to
    target program type, and check equivalence.
    """
    source_circuit = circuits[circuit_name]
    circuit = transpile(source_circuit, "cirq", conversion_graph=graph)
    qasm = circuit.to_qasm()
    cirq_circuit = circuit_from_qasm(qasm)
    target_circuit = transpile(cirq_circuit, target)
    assert circuits_allclose(cirq_circuit, target_circuit, strict_gphase=False)


@pytest.mark.parametrize(("target", "baseline"), AVAILABLE_TARGETS)
def test_pytket_coverage(target, baseline, pytket_circuits, conversion_graph):
    """Test converting PyTKET circuits to supported target program type over
    all PyTKET gates and check against baseline expecte accuracy.
    """
    ACCURACY_BASELINE = baseline
    ALLOWANCE = 0.01
    failures = {}
    for gate_name in pytket_circuits:
        try:
            convert_from_pytket_to_x(target, gate_name, pytket_circuits, conversion_graph)
        except Exception as e:  # pylint: disable=broad-exception-caught
            failures[f"{target}-{gate_name}"] = e

    total_tests = len(pytket_circuits)
    nb_fails = len(failures)
    nb_passes = total_tests - nb_fails
    accuracy = float(nb_passes) / float(total_tests)

    assert accuracy >= ACCURACY_BASELINE - ALLOWANCE, (
        f"The coverage threshold was not met. {nb_fails}/{total_tests} tests failed "
        f"({nb_fails / (total_tests):.2%}) and {nb_passes}/{total_tests} passed "
        f"(expected >= {ACCURACY_BASELINE}).\nFailures: {failures.keys()}\n\n"
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Benchmarking tests for Qiskit conversions

"""
import importlib.util

import pytest
import qiskit

from qbraid.interface import circuits_allclose
from qbraid.transpiler import ConversionGraph, transpile

from ...fixtures.qiskit.gates import get_qiskit_gates


def is_package_installed(package_name: str) -> bool:
    """Check if a package is installed."""
    return importlib.util.find_spec(package_name) is not None


ALL_TARGETS = [("braket", 1.0), ("cirq", 1.0), ("pyquil", 0.98), ("pytket", 1.0)]
AVAILABLE_TARGETS = [(name, version) for name, version in ALL_TARGETS if is_package_installed(name)]

qiskit_gates = get_qiskit_gates(seed=0, exclude=["GlobalPhaseGate"])

graph = ConversionGraph(require_native=True)


def convert_from_qiskit_to_x(target, gate_name):
    """Construct a Qiskit circuit with the given gate, transpile it to
    target program type, and check equivalence.
    """
    gate = qiskit_gates[gate_name]
    source_circuit = qiskit.QuantumCircuit(gate.num_qubits)
    source_circuit.compose(gate, inplace=True)
    target_circuit = transpile(source_circuit, target, conversion_graph=graph)
    assert circuits_allclose(source_circuit, target_circuit, strict_gphase=False)


@pytest.mark.parametrize(("target", "baseline"), AVAILABLE_TARGETS)
def test_qiskit_coverage(target, baseline):
    """Test converting Qiskit circuits to supported target program type over
    all Qiskit standard gates and check against baseline expected accuracy.
    """
    ACCURACY_BASELINE = baseline
    ALLOWANCE = 0.01
    failures = {}
    for gate_name in qiskit_gates:
        try:
            convert_from_qiskit_to_x(target, gate_name)
        except Exception as e:  # pylint: disable=broad-exception-caught
            failures[f"{target}-{gate_name}"] = e

    total_tests = len(qiskit_gates)
    nb_fails = len(failures)
    nb_passes = total_tests - nb_fails
    accuracy = float(nb_passes) / float(total_tests)

    assert accuracy >= ACCURACY_BASELINE - ALLOWANCE, (
        f"The coverage threshold was not met. {nb_fails}/{total_tests} tests failed "
        f"({nb_fails / (total_tests):.2%}) and {nb_passes}/{total_tests} passed "
        f"(expected >= {ACCURACY_BASELINE}).\nFailures: {failures.keys()}\n\n"
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining CudaQKernel Class

"""
from __future__ import annotations

import base64

import cudaq

from qbraid.programs.exceptions import ProgramTypeError

from ._model import GateModelProgram


class CudaQKernel(GateModelProgram):
    """Wrapper class for ``cudaq.PyKernel`` objects."""

    def __init__(self, program: cudaq.PyKernel):
        super().__init__(program)
        if not isinstance(program, cudaq.PyKernel):
            raise ProgramTypeError(
                message=f"Expected 'cudaq.PyKernel' object, got '{type(program)}'."
            )

    @property
    def qubits(self) -> list[int]:
        """Return the qubits acted upon by the operations in this circuit"""
        return list(range(self.num_qubits))

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        state = cudaq.get_state(self.program)
        return state.num_qubits()

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return 0

    def serialize(self) -> dict[str, str]:
        """Return the program in a format suitable for submission to the qBraid API."""
        qir: str = cudaq.translate(self.program, format="qir-base")
        return {"qir": base64.b64encode(qir.encode("utf-8")).decode("utf-8")}



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for converting Braket circuits to/from OpenQASM 3

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import pyqasm
from qbraid_core._import import LazyLoader

from qbraid.passes.qasm.compat import (
    convert_qasm_pi_to_decimal,
    remove_stdgates_include,
    replace_gate_names,
)
from qbraid.programs.exceptions import QasmError
from qbraid.transpiler.annotations import weight

braket_circuits = LazyLoader("braket_circuits", globals(), "braket.circuits")
braket_openqasm = LazyLoader("braket_openqasm", globals(), "braket.ir.openqasm")

if TYPE_CHECKING:
    import braket.circuits

    from qbraid.programs.typer import Qasm3StringType


def transform_notation(qasm3: str) -> str:
    """
    Process an OpenQASM 3 program that was generated by
    an external tool to make it compatible with Amazon Braket.

    """
    replacements = {
        "cx": "cnot",
        "sdg": "si",
        "tdg": "ti",
        "sx": "v",
        "id": "i",
        "sxdg": "vi",
        "p": "phaseshift",
        "cp": "cphaseshift",
    }

    qasm3 = remove_stdgates_include(qasm3)
    qasm3 = replace_gate_names(qasm3, replacements)
    qasm3 = convert_qasm_pi_to_decimal(qasm3)
    return qasm3


@weight(1)
def qasm3_to_braket(qasm: Qasm3StringType) -> braket.circuits.Circuit:
    """Converts an OpenQASM 3.0 string to a ``braket.circuits.Circuit``.

    Args:
        qasm: OpenQASM 3 string

    Returns:
        The Amazon Braket circuit equivalent to the input OpenQASM 3.0 string

    Raises:
        ProgramConversionError: If qasm to braket conversion fails

    """
    module = pyqasm.loads(qasm)
    module.unroll()
    qasm = pyqasm.dumps(module)

    qasm = transform_notation(qasm)

    try:
        program = braket_openqasm.Program(source=qasm)
        return braket_circuits.Circuit.from_ir(source=program.source, inputs=program.inputs)
    except Exception as err:
        raise QasmError("Error converting qasm3 string to braket circuit") from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.pennylane.PennylaneTape

"""
import numpy as np
import pennylane as qml
import pytest
from pennylane.tape import QuantumTape

from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.gate_model.pennylane import PennylaneTape


def two_wire_tape(wire1, wire2):
    """Returns a simple Pennylane tape with two wires."""
    with QuantumTape() as tape:
        qml.Hadamard(wires=wire1)
        qml.CNOT(wires=[wire1, wire2])
    return tape


@pytest.mark.parametrize("bell_circuit", ["pennylane"], indirect=True)
def test_pennylane_tape(bell_circuit, bell_unitary):
    """Test PennylaneTape class."""
    tape, _ = bell_circuit
    program = PennylaneTape(tape)
    assert isinstance(program.program, QuantumTape)
    assert program.qubits == [0, 1]
    assert program.num_qubits == 2
    assert program.num_clbits == 0
    assert program.depth == 2
    assert np.allclose(program.unitary(), bell_unitary)


@pytest.mark.parametrize("wires", [(0, 1), ("a", "b")])
def test_pennylane_reverse_qubit_order(wires):
    """Test PennylaneTape class with qubits in reverse order."""
    w1, w2 = wires

    tape_in = two_wire_tape(w1, w2)
    tape_expected = two_wire_tape(w2, w1)

    wires_in = [op.wires for op in tape_in.operations]
    wires_expected = [op.wires for op in tape_expected.operations]

    assert wires_in != wires_expected

    program = PennylaneTape(tape_in)
    program.reverse_qubit_order()
    tape_out = program.program

    wires_out = [op.wires for op in tape_out.operations]

    assert wires_out == wires_expected
    assert str(tape_out) == str(tape_expected)
    assert repr(tape_out) == repr(tape_expected)
    assert qml.equal(tape_out, tape_expected)


@pytest.mark.parametrize("wires", [(3, 6), (2, 4)])
def test_pennylane_remove_idle_qubits(wires):
    """Test PennylaneTape class with qubits in reverse order."""
    tape_in = two_wire_tape(*wires)
    program = PennylaneTape(tape_in)
    program.remove_idle_qubits()
    tape_out = program.program

    wires_out = [op.wires.tolist() for op in tape_out.operations]

    assert wires_out == [[0], [0, 1]]
    assert tape_out.wires.toset() == {0, 1}


def test_raise_program_type_error():
    """Test raising ProgramTypeError"""
    with pytest.raises(ProgramTypeError):
        PennylaneTape("OPENQASM 2.0;qreg q[2];h q[0];cx q[0],q[1];")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=unused-argument,redefined-outer-name
"""
Unit tests for plotting conversion graphs

"""
import warnings
from unittest.mock import MagicMock, Mock, patch

import pytest

from qbraid.programs.experiment import ExperimentType
from qbraid.programs.spec import ProgramSpec
from qbraid.programs.typer import QuboCoefficientsDict
from qbraid.runtime.native import QbraidDevice
from qbraid.runtime.profile import TargetProfile
from qbraid.transpiler.edge import Conversion
from qbraid.transpiler.graph import ConversionGraph
from qbraid.visualization.plot_conversions import (
    _validate_colors,
    plot_conversion_graph,
    plot_runtime_conversion_scheme,
)


@pytest.fixture
def mock_graph():
    """Mock graph object for testing plot_conversion_graph."""
    mock = MagicMock()
    mock.nodes.return_value = ["node1", "node2"]
    mock.conversions.return_value = []
    mock.edge_list.return_value = []
    mock.get_node_data.return_value = "node_data"
    mock.get_edge_data.return_value = {"native": True}
    mock._node_alias_id_map = {"node1": 1, "node2": 2}
    return mock


@patch("qbraid.visualization.plot_conversions.plt", autospec=True)
@patch(
    "qbraid.visualization.plot_conversions.rx.spring_layout",
    return_value={"node1": (0, 0), "node2": (1, 1)},
    autospec=True,
)
def test_plot_conversion_graph_show(mock_layout, mock_plt, mock_graph):
    """Test that the graph is displayed when show is True and not saved."""
    plot_conversion_graph(graph=mock_graph, show=True, edge_labels=True)
    mock_plt.show.assert_called_once()
    mock_plt.savefig.assert_not_called()


@patch("qbraid.visualization.plot_conversions.plt", autospec=True)
@patch("qbraid.visualization.plot_conversions.rx.spring_layout", autospec=True)
def test_plot_conversion_graph_save(mock_layout, mock_plt, mock_graph):
    """Test that the graph is saved to the correct path when specified."""
    save_path = "path/to/save/plot.png"
    plot_conversion_graph(graph=mock_graph, show=False, save_path=save_path)
    mock_plt.savefig.assert_called_once_with(save_path, bbox_inches="tight", dpi=300)
    mock_plt.show.assert_not_called()


@pytest.fixture
def mock_annealer_profile():
    """Mock profile for testing."""
    return TargetProfile(
        device_id="mock_annealer",
        simulator=True,
        experiment_type=ExperimentType.ANNEALING,
        num_qubits=42,
        program_spec=ProgramSpec(QuboCoefficientsDict, alias="qubo"),
        noise_models=None,
    )


@pytest.fixture
def mock_annealer_device(mock_annealer_profile):
    """Mock device for testing."""
    return QbraidDevice(mock_annealer_profile, Mock())


@patch("qbraid.visualization.plot_conversions.plt", autospec=True)
@patch("qbraid.visualization.plot_conversions.mpl_draw", autospec=True)
def test_plot_runtime_conversion_scheme_qubo_node_target(mock_draw, mock_plt, mock_annealer_device):
    """Test plotting runtime conversion scheme for device with qubo target."""
    plot_runtime_conversion_scheme(mock_annealer_device, legend=True, show=False)
    mock_plt.title.assert_called_once_with(
        "Runtime Conversion Scheme for QbraidDevice('mock_annealer')"
    )
    mock_draw.assert_called_once()
    _, kwargs = mock_draw.call_args
    assert "edgecolors" in kwargs
    assert isinstance(kwargs["edgecolors"], list)
    assert len(kwargs["edgecolors"]) == 1
    assert kwargs["edgecolors"][0] == "red"

    mock_plt.legend.assert_called_once()
    _, kwargs = mock_plt.legend.call_args
    assert "handles" in kwargs
    assert isinstance(kwargs["handles"], list)
    assert len(kwargs["handles"]) == 2


@patch("qbraid.visualization.plot_conversions.plt", autospec=True)
def test_invoke_plot_method_from_conversion_graph(mock_plt):
    """Test that the graph is displayed when invoked via ConversionGraph.plot method."""
    with patch("rustworkx.__version__", "0.15.1"):
        graph = ConversionGraph()

        with pytest.warns(UserWarning, match=r"Detected multiple edge colors*"):
            graph.plot(show=True)
            mock_plt.show.assert_called_once()


def test_plot_conversion_graph_warning():
    """Test that a warning is raised when multiple edge colors are detected."""
    with patch("rustworkx.__version__", "0.15.0"):
        graph = ConversionGraph()

        graph.add_conversion(Conversion("node1", "node2", lambda x: x))

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            plot_conversion_graph(graph, show=False)
            assert any(
                "Detected multiple edge colors, which may not display correctly"
                in str(warning.message)
                for warning in w
            )


def test_plot_conversion_graph_experiment_type_raises_for_none_found():
    """Test that an exception is raised when no matching experiment type is found."""
    graph = ConversionGraph(nodes=["qasm2", "qasm3"])

    with pytest.raises(ValueError) as exc:
        plot_conversion_graph(graph, experiment_type=ExperimentType.ANNEALING)
    assert "No program type nodes found" in str(exc.value)


@patch("qbraid.visualization.plot_conversions.is_registered_alias_native", autospec=True)
def test_plot_conversion_graph_valid_experiment_type(is_registered_alias_native):
    """Test plotting a conversion graph based on a valid experiment type."""
    nodes = ["qasm2", "qasm3", "qubo"]
    conversions = [Conversion("qasm2", "qasm3", lambda x: x)]
    graph = ConversionGraph(conversions=conversions, nodes=nodes, include_isolated=True)

    plot_conversion_graph(graph, experiment_type=ExperimentType.ANNEALING, show=False)
    is_registered_alias_native.assert_called_once_with("qubo")


@patch("qbraid.visualization.plot_conversions.plt", autospec=True)
@patch("qbraid.visualization.plot_conversions.rx.spring_layout", autospec=True)
def test_plot_conversion_graph_with_custom_title_and_legend(mock_layout, mock_plt, mock_graph):
    """Test custom title and legend are set correctly."""
    title = "Custom Graph Title"
    plot_conversion_graph(graph=mock_graph, title=title, legend=True, show=False)
    mock_plt.title.assert_called_once_with(title)
    mock_plt.legend.assert_called_once()


def test_validate_colors_with_custom_input():
    """Test _validate_colors when colors is not None."""
    custom_colors = {
        "target_node_outline": "green",
        "qbraid_node": "yellow",
    }

    expected_colors = {
        "target_node_outline": "green",
        "qbraid_node": "yellow",
        "external_node": "lightgray",
        "qbraid_edge": "gray",
        "external_edge": "blue",
        "extras_edge": "darkgrey",
    }

    result = _validate_colors(custom_colors)
    assert result == expected_colors


def test_validate_colors_with_unexpected_keys():
    """Test _validate_colors raises ValueError for unexpected keys."""
    invalid_colors = {"unexpected_key": "purple"}

    with pytest.raises(ValueError, match="Unexpected keys in 'colors': {'unexpected_key'}"):
        _validate_colors(invalid_colors)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for drawing quantum circuit diagrams

"""
from __future__ import annotations

import importlib.util
from typing import TYPE_CHECKING, Any, Optional

from pyqasm.printer import draw

from qbraid.programs import QPROGRAM_ALIASES, ProgramTypeError, get_program_type_alias
from qbraid.transpiler.converter import transpile
from qbraid.transpiler.exceptions import ConversionPathNotFoundError

from .exceptions import VisualizationError

if TYPE_CHECKING:
    import qbraid.programs


def circuit_drawer(
    program: qbraid.programs.QPROGRAM,
    as_package: Optional[str] = None,
    output: Optional[str] = None,
    **kwargs,
) -> Any:
    """Draws circuit diagram.

    Args:
        program: Supported quantum program
        as_package: The package to convert the program to before drawing
        output: The output format for the circuit diagram

    Raises:
        ProgramTypeError: If quantum program is not of a supported type
        ValueError: If an invalid output option is provided
    """
    package = get_program_type_alias(program)

    if as_package and as_package != package:
        if as_package not in QPROGRAM_ALIASES:
            raise ValueError(
                f"Invalid package '{as_package}'. "
                "Make sure the desired output package is installed."
            )
        program = transpile(program, as_package)
        package = as_package

    # pylint: disable=import-outside-toplevel

    if package == "qiskit":
        from qiskit.visualization import circuit_drawer as qiskit_drawer

        return qiskit_drawer(program, output=output, **kwargs)

    if package == "braket":
        if output in {None, "ascii"}:
            from braket.circuits.ascii_circuit_diagram import AsciiCircuitDiagram

            return print(AsciiCircuitDiagram.build_diagram(program))
        raise ValueError('The only valid option for braket are "ascii"')

    if package == "cirq":
        if output in {None, "text"}:
            return print(program.to_text_diagram(**kwargs))  # type: ignore[attr-defined]
        if output == "svg":
            from cirq.contrib.svg import SVGCircuit

            # pragma: no cover
            return SVGCircuit(program)
        if output == "svg_source":
            from cirq.contrib.svg import circuit_to_svg

            return circuit_to_svg(program)
        raise ValueError('The only valid options for cirq are "text", "svg", "svg_source"')

    if package == "pyquil":
        if output is None or output == "text":
            return print(program)
        if output == "latex":
            from pyquil.latex import display

            return display(program, **kwargs)  # pragma: no cover
        raise ValueError('The only valid options for pyquil are "text", "latex"')

    if package == "pytket":
        if output in {None, "jupyter"}:
            from pytket.circuit.display import render_circuit_jupyter

            return render_circuit_jupyter(program)  # pragma: no cover
        if output == "view_browser":
            from pytket.circuit.display import view_browser

            return view_browser(program, **kwargs)  # pragma: no cover
        if output == "html":
            from pytket.circuit.display import render_circuit_as_html

            return render_circuit_as_html(program, **kwargs)  # pragma: no cover
        raise ValueError('The only valid options for pytket are "jupyter", "view_browser", "html"')

    if package in {"qasm2", "qasm3"}:
        if output == "mpl" or (
            output is None and importlib.util.find_spec("matplotlib") is not None
        ):
            return draw(program, output="mpl", **kwargs)

        try:
            if package == "qasm2" and "cirq" in QPROGRAM_ALIASES:
                program = transpile(program, "cirq")
            elif "qiskit" in QPROGRAM_ALIASES:
                program = transpile(program, "qiskit")
            elif "braket" in QPROGRAM_ALIASES:
                program = transpile(program, "braket")
            else:
                raise ValueError("No supported package found for drawing circuit")
        except ConversionPathNotFoundError as err:
            raise VisualizationError(
                "Unable to convert program to a supported package for drawing"
            ) from err

        return circuit_drawer(program, output=output, **kwargs)

    if package == "pennylane":
        program = transpile(program, "qasm2")

        return circuit_drawer(program, output=output, **kwargs)

    raise ProgramTypeError(package)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Oxford Quantum Circuits (OQC) provider class

"""
import json
import os
from typing import Any, Optional, Union

from qcaas_client.client import OQCClient

from qbraid._caching import cached_method
from qbraid.programs import ExperimentType
from qbraid.programs.spec import ProgramSpec
from qbraid.runtime.exceptions import ResourceNotFoundError
from qbraid.runtime.profile import TargetProfile
from qbraid.runtime.provider import QuantumProvider
from qbraid.runtime.schemas.base import USD

from .device import OQCDevice

NATIVE_GATES = {"qpu:jp:3:673b1ad43c": {"ecr", "id", "rz", "sx", "x"}}


class OQCProvider(QuantumProvider):
    """OQC provider class."""

    def __init__(
        self,
        token: Optional[str] = None,
        url: str = "https://cloud.oqc.app/",
        timeout: tuple[int, int] = (5, 10),
    ) -> None:
        token = token or os.getenv("OQC_AUTH_TOKEN")
        if not token:
            raise ValueError(
                "An OQC authenication token is required to initialize the provider. "
                "Please provide it directly as an argument or set it via "
                "the OQC_AUTH_TOKEN environment variable."
            )
        self.client = OQCClient(url=url, authentication_token=token, timeout=timeout)

    def _build_profile(self, data: dict[str, Any]) -> TargetProfile:
        """Build a profile for OQC device."""
        data = data.copy()
        device_id = data.get("id")
        feature_set: Union[str, dict] = data.get("feature_set", {})

        if not isinstance(feature_set, dict):
            try:
                data["feature_set"] = json.loads(feature_set)
            except json.JSONDecodeError as err:
                raise ValueError(
                    f"Failed to decode feature set data for device '{device_id}'."
                ) from err

        try:
            device_id: str = data.pop("id")
            device_name: str = data.pop("name")
            endpoint_url: str = data.pop("url")
            feauture_set: dict = data.pop("feature_set")
            num_qubits: int = feauture_set.pop("qubit_count")
            simulator: bool = feauture_set.pop("simulator", False)
        except KeyError as err:
            raise ValueError(f"Failed to gather profile data for device '{device_id}'.") from err

        dynamic_fields = {"active", "status"}
        pricing_fields = {"price_per_shot", "price_per_task"}
        static_data = {
            k: (USD(v) if k in pricing_fields else v)
            for k, v in data.items()
            if k not in dynamic_fields
        }

        return TargetProfile(
            device_id=device_id,
            simulator=simulator,
            experiment_type=ExperimentType.GATE_MODEL,
            num_qubits=num_qubits,
            program_spec=ProgramSpec(str, alias="qasm3"),
            device_name=device_name,
            endpoint_url=endpoint_url,
            provider_name="OQC",
            feature_set=feauture_set,
            **static_data,
        )

    @cached_method
    def get_devices(self) -> list[OQCDevice]:
        """Get all OQC devices."""
        devices: list[dict] = self.client.get_qpus()
        return [
            OQCDevice(profile=self._build_profile(device), client=self.client) for device in devices
        ]

    @cached_method
    def get_device(self, device_id: str) -> OQCDevice:
        """Get a specific OQC device."""
        devices: list[dict] = self.client.get_qpus()
        device = next((d for d in devices if d["id"] == device_id), None)
        if not device:
            raise ResourceNotFoundError(f"Device '{device_id}' not found.")
        return OQCDevice(profile=self._build_profile(device), client=self.client)

    def __hash__(self):
        if not hasattr(self, "_hash"):
            object.__setattr__(
                self, "_hash", hash((self.client.url, self.client._authentication_token))
            )
        return self._hash  # pylint: disable=no-member



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining BraketQuantumTask Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import boto3
from braket.aws import AwsQuantumTask
from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import (
    AnalogHamiltonianSimulationQuantumTaskResult,
)
from braket.tasks.gate_model_quantum_task_result import GateModelQuantumTaskResult

from qbraid._logging import logger
from qbraid.runtime.enums import JobStatus
from qbraid.runtime.exceptions import JobStateError
from qbraid.runtime.job import QuantumJob
from qbraid.runtime.result import Result
from qbraid.runtime.result_data import AhsResultData, GateModelResultData

from .result_builder import BraketAhsResultBuilder, BraketGateModelResultBuilder
from .tracker import get_quantum_task_cost

if TYPE_CHECKING:
    from decimal import Decimal

    from braket.aws import AwsSession

AWS_TASK_STATUS_MAP = {
    "CREATED": JobStatus.INITIALIZING,
    "QUEUED": JobStatus.QUEUED,
    "RUNNING": JobStatus.RUNNING,
    "CANCELLING": JobStatus.CANCELLING,
    "CANCELLED": JobStatus.CANCELLED,
    "COMPLETED": JobStatus.COMPLETED,
    "FAILED": JobStatus.FAILED,
}


class AmazonBraketVersionError(Exception):
    """Exception raised for Amazon Braket SDK errors due to versioning."""


class BraketQuantumTask(QuantumJob):
    """Wrapper class for Amazon Braket ``QuantumTask`` objects."""

    def __init__(self, task_id: str, task: Optional[AwsQuantumTask] = None, **kwargs):
        """Create a BraketQuantumTask."""

        super().__init__(task_id, **kwargs)
        self._task = task or AwsQuantumTask(task_id)

    def status(self):
        """Returns status from Braket QuantumTask object metadata."""
        state = self._task.state()
        status = AWS_TASK_STATUS_MAP.get(state, JobStatus.UNKNOWN)
        self._cache_metadata["status"] = status
        return status

    def queue_position(self) -> Optional[int]:
        """Returns queue position from Braket QuantumTask.
        '>2000' returns as 2000 for typing consistency."""
        try:
            position = self._task.queue_position().queue_position
            if isinstance(position, str):
                if position.startswith(">"):
                    position = position[1:]
                return int(position)
            return position
        except AttributeError as err:
            raise AmazonBraketVersionError(
                "Queue visibility is only available for amazon-braket-sdk>=1.56.0"
            ) from err

    def result(self) -> Result:
        """Return the results of the job."""
        if not self.is_terminal_state():
            logger.info("Result will be available when the job has reached a final state.")

        bk_result = self._task.result()
        metadata = self._task.metadata()
        success = metadata["status"] == "COMPLETED"
        device_id = metadata["deviceArn"]
        job_id = metadata["quantumTaskArn"]

        result_mapping = {
            GateModelQuantumTaskResult: (BraketGateModelResultBuilder, GateModelResultData),
            AnalogHamiltonianSimulationQuantumTaskResult: (BraketAhsResultBuilder, AhsResultData),
        }

        builder_class, data_class = result_mapping.get(type(bk_result), (None, None))

        if not builder_class or not data_class:
            raise ValueError(f"Unsupported result type: {type(bk_result).__name__}")

        # Retrieve partial measurement qubit information from job tags
        partial_measurement_qubits = self._get_partial_measurement_qubits_from_tags(
            bk_result.measured_qubits
        )
        result_data = {
            "measurement_counts": (
                builder_class(bk_result, partial_measurement_qubits).get_counts()
                if success
                else None
            ),
            "measurements": (
                builder_class(bk_result, partial_measurement_qubits).measurements()
                if success
                else None
            ),
        }

        data = data_class(**result_data)
        return Result(device_id=device_id, job_id=job_id, success=success, data=data, **metadata)

    def cancel(self) -> None:
        """Cancel the quantum task."""
        task = self._task

        if self.is_terminal_state():
            raise JobStateError("Cannot cancel quantum job in terminal state.")

        try:
            task.cancel()
        except RuntimeError:
            task._aws_session.cancel_quantum_task(self.id)

    @staticmethod
    def _get_cost(task_arn: str, aws_session: Optional[AwsSession] = None) -> Decimal:
        """Return the cost of the quantum task."""
        return get_quantum_task_cost(task_arn, aws_session=aws_session)

    def get_cost(self) -> float:
        """Return the cost of the job."""
        decimal_cost = self._get_cost(self.id)
        return float(decimal_cost)

    def _get_partial_measurement_qubits_from_tags(
        self, all_measurement_qubits: list[int]
    ) -> list[int] | None:
        """
        Retrieve partial measurement qubit indices from quantum task tags.

        This method queries the AWS Braket service to get the quantum task metadata
        and extracts the partial measurement qubit information that was stored as tags
        during task submission. It then maps these qubit indices to their positions
        in the measurement results array.

        Args:
            all_measurement_qubits: List of all qubits that were measured in the circuit,
                in the order they appear in the measurement results.

        Returns:
            List of indices corresponding to the positions of partial measurement qubits
            in the measurement results array, or None if no partial measurements were used.
        """
        braket_client = boto3.client("braket", region_name=self._task._aws_session.region)
        response = braket_client.get_quantum_task(quantumTaskArn=self._task.id)

        if "partial_measurement_qubits" not in response["tags"]:
            return None

        # Parse the partial measurement qubit indices from the tag string (e.g., "0/2/3")
        partial_measurement_qubits_str = response["tags"]["partial_measurement_qubits"]
        partial_measurement_qubits = [int(q) for q in partial_measurement_qubits_str.split("/")]

        # Map the original qubit indices to their positions in the measurement results array
        partial_measurement_qubit_indices = [
            all_measurement_qubits.index(q) for q in partial_measurement_qubits
        ]
        return partial_measurement_qubit_indices



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for defining custom conversions

"""
from __future__ import annotations

import importlib.util
import inspect
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

import numpy as np

from qbraid.programs import QPROGRAM_REGISTRY, get_program_type_alias

if TYPE_CHECKING:
    import qbraid.programs


class Conversion:
    """
    Class for defining and handling custom conversions between different quantum program packages.

    """

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        source: str,
        target: str,
        conversion_func: Callable,
        weight: Optional[float] = None,
        bias: Optional[float] = None,
    ):
        """
        Initialize a Conversion instance with source and target packages and a conversion function.

        Args:
            source (str): The source package from which conversion starts.
            target (str): The target package to which conversion is done.
            conversion_func (Callable): The function that performs the actual conversion.
            weight (Optional[float]): Optional weighting factor for the conversion, ranging [0,1].
                If not specified, defaults to 1 or a custom value derived from the conversion_func.
            bias (Optional[float]): Optional factor used to fine-tune the weight calculation and
                modify the decision thresholds for pathfinding. Defaults to 0. Higher values
                prioritize shorter paths. For example, a bias of 0.25 slightly favors a single
                conversion at weight 0.8 over two conversions at weight 1.0, whereas a bias of 0.1
                requires a single conversion of weight > 0.9 to be preferred over two at weight 1.0.
        """
        self._source = source
        self._target = target
        self._conversion_func = conversion_func
        self._bias = bias if bias is not None else 0
        self._weight = self._get_adjusted_weight(weight)
        self._extras = getattr(conversion_func, "requires_extras", [])
        self._native = self._is_module_native(conversion_func)
        self._supported = self._is_conversion_supported()

    @property
    def source(self) -> str:
        """
        The source package of the conversion.

        Returns:
            str: The source package name.
        """
        return self._source

    @property
    def target(self) -> str:
        """
        The target package of the conversion.

        Returns:
            str: The target package name.
        """
        return self._target

    @property
    def native(self) -> bool:
        """
        True if the conversion function is native to qbraid package, False otherwise.

        Returns:
            bool: Whether the conversion function is native to qbraid package.
        """
        return self._native

    @property
    def supported(self) -> bool:
        """
        True if all packages required to perform the conversion are installed. False otherwise.

        Returns:
            bool: Whether the conversion function supported in the current runtime environment.
        """
        return self._supported

    @property
    def weight(self) -> int:
        """
        The weight of the conversion function used to prioritize conversion paths.

        Returns:
            int: The weight of the conversion function.
        """
        return self._weight

    def _get_adjusted_weight(self, weight: Optional[float] = None) -> float:
        """
        Calculates and returns the effective weight of the conversion, applying a bias to
        prioritize shorter conversion paths when used with pathfinding algorithms like rustworkx.

        Args:
            weight (float, optional): The initial weight provided by the user. Defaults to
                the weight attribute of conversion_func if not provided.

        Returns:
            float: The calculated weight adjusted for pathfinding optimization.

        Raises:
            ValueError: If the calculated or provided weight is not between 0 and 1, inclusive.
        """
        effective_weight = (
            weight if weight is not None else getattr(self._conversion_func, "weight", 1)
        )

        if not 0 <= effective_weight <= 1:
            raise ValueError("Weight must be a float between 0 and 1, inclusive.")

        # Invert and log transform for positive weight differentiation with rustworkx
        rx_adjusted_weight = float("inf") if effective_weight == 0 else np.log(1 / effective_weight)

        adjusted_weight = rx_adjusted_weight + self._bias

        return adjusted_weight

    def _is_module_native(self, func: Callable) -> bool:
        """
        Determine if the function's module is 'qbraid' and requires no extras.

        Args:
            func (Callable): The function to check the module of.

        Returns:
            bool: True if the module is 'qbraid' and requires no extras, False otherwise.
        """
        module = inspect.getmodule(func)
        is_native = (
            module is not None
            and module.__name__.split(".")[0] == "qbraid"
            and len(self._extras) == 0
            and getattr(func, "weight", None) is not None
        )
        return is_native

    def _is_conversion_supported(self) -> bool:
        """
        Determine if the required packages for the conversion are installed.

        Returns:
            bool: True if supported, otherwise False.
        """
        if self._native:
            return True
        return all(importlib.util.find_spec(m) is not None for m in self._extras)

    def convert(self, program: qbraid.programs.QPROGRAM) -> Union[qbraid.programs.QPROGRAM, Any]:
        """
        Convert a quantum program from the source package to the target package.

        Args:
            program (qbraid.programs.QPROGRAM): The quantum program to be converted.

        Returns:
            Union[qbraid.programs.QPROGRAM, Any]: The converted quantum program,
                                         typically of a supported program type.

        Raises:
            ValueError: If the provided program's type does not match the source package type.
        """
        package = get_program_type_alias(program)
        if package != self._source:
            raise ValueError(
                f"Expected program of type {QPROGRAM_REGISTRY[self._source]}, "
                f"but got program of type {QPROGRAM_REGISTRY[package]}."
            )
        return self._conversion_func(program)

    def __repr__(self) -> str:
        """
        Represent the Conversion instance as a string indicating
        source and target packages.

        Returns:
            str: String representation of the Conversion instance.
        """
        return f"('{self._source}', '{self._target}')"

    def __eq__(self, other: Any) -> bool:
        """
        Check if another instance is equal to this instance.

        Args:
            other (Any): Another instance to compare.

        Returns:
            bool: True if the instances are equal, False otherwise.
        """
        if not isinstance(other, Conversion):
            return False

        return (
            self._source == other._source
            and self._target == other._target
            and self._native == other._native
            and self._supported == other._supported
            and self._extras == other._extras
            and self._weight == other._weight
        )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM conversion function

"""
import pyqasm

from qbraid.programs.typer import Qasm2String, Qasm2StringType, Qasm3StringType
from qbraid.transpiler.annotations import weight


@weight(0.7)
def qasm2_to_qasm3(qasm_str: Qasm2StringType) -> Qasm3StringType:
    """Convert a OpenQASM 2.0 string to OpenQASM 3.0 string

    Args:
        qasm_str (str): OpenQASM 2.0 string

    Returns:
        str: OpenQASM 3.0 string
    """
    if not isinstance(qasm_str, Qasm2String):
        raise ValueError("Invalid OpenQASM 2.0 string")

    qasm_module = pyqasm.loads(qasm_str)
    return qasm_module.to_qasm3(as_str=True)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining AnalogHamiltonianProgram Class

"""
from __future__ import annotations

from decimal import Decimal
from typing import TYPE_CHECKING, Union

import braket.ir.ahs as ir
from braket.ahs import AnalogHamiltonianSimulation, Field

from qbraid.programs.exceptions import ProgramTypeError

from ._model import AnalogHamiltonianProgram

if TYPE_CHECKING:
    import braket.ahs

    import qbraid.runtime.aws


class BraketAHS(AnalogHamiltonianProgram):
    """Wrapper class for ``braket.ahs.AnalogHamiltonianSimulation`` objects."""

    def __init__(self, program: braket.ahs.AnalogHamiltonianSimulation):
        super().__init__(program)
        if not isinstance(program, AnalogHamiltonianSimulation):
            raise ProgramTypeError(
                message=f"Expected 'ahs.AnalogHamiltonianSimulation' object, got '{type(program)}'."
            )

    def to_dict(self) -> dict:
        return BraketAHSEncoder().encode_ahs(self.program)

    def transform(self, device: qbraid.runtime.aws.BraketDevice) -> None:
        """Transform program to according to device target profile."""
        if not device.simulator:
            program: AnalogHamiltonianSimulation = self.program
            self._program = program.discretize(device._device)


class BraketAHSEncoder:
    """Class for encoding AnalogHamiltonianSimulation objects to dictionaries."""

    def encode_register(self, register: ir.AtomArrangement) -> dict:
        """Convert an AtomArrangement object to a dictionary."""
        return {
            "sites": [[str(value) for value in sublist] for sublist in register.sites],
            "filling": register.filling,
        }

    def encode_field(self, field: Union[ir.PhysicalField, Field]) -> dict:
        """Convert a PhysicalField object to a dictionary."""
        return {
            "time_series": {
                "values": [str(v) for v in field.time_series.values],
                "times": [str(t) for t in field.time_series.times],
            },
            "pattern": field.pattern,
        }

    def encode_hamiltonian(self, hamiltonian: ir.Hamiltonian) -> dict:
        """Convert a Hamiltonian object to a dictionary."""
        return {
            "drivingFields": [
                {
                    "amplitude": self.encode_field(field.amplitude),
                    "phase": self.encode_field(field.phase),
                    "detuning": self.encode_field(field.detuning),
                }
                for field in hamiltonian.drivingFields
            ],
            "localDetuning": [
                {
                    "magnitude": self.encode_field(term.magnitude),
                }
                for term in hamiltonian.localDetuning
            ],
        }

    def encode_ahs(self, program: AnalogHamiltonianSimulation) -> dict:
        """Convert an AnalogHamiltonianSimulation object to a dictionary."""
        program_ir = program.to_ir()

        return {
            "register": self.encode_register(program_ir.setup.ahs_register),
            "hamiltonian": self.encode_hamiltonian(program_ir.hamiltonian),
        }


class BraketAHSDecoder:
    """Class for decoding AnalogHamiltonianSimulation objects from dictionaries."""

    def decode_time_series(self, values: list[str], times: list[str]) -> ir.TimeSeries:
        """Create a TimeSeries object from lists of values and times."""
        if len(values) != len(times):
            raise ValueError("The values and times lists must have the same length.")
        return ir.TimeSeries(values=[Decimal(v) for v in values], times=[Decimal(t) for t in times])

    def decode_physical_field(
        self, values: list[str], times: list[str], pattern: Union[str, list[float]]
    ) -> ir.PhysicalField:
        """Create a PhysicalField object from lists of values and times."""
        time_series = self.decode_time_series(values, times)
        return ir.PhysicalField(time_series=time_series, pattern=pattern)

    def decode_driving_field(self, amplitude_info, phase_info, detuning_info) -> ir.DrivingField:
        """
        Create a DrivingField object from lists of values and times for
        the amplitude, phase, and detuning.

        """
        amplitude = self.decode_physical_field(*amplitude_info)
        phase = self.decode_physical_field(*phase_info)
        detuning = self.decode_physical_field(*detuning_info)
        return ir.DrivingField(amplitude=amplitude, phase=phase, detuning=detuning)

    def decode_local_detuning(self, magnitude_info) -> ir.LocalDetuning:
        """
        Create a LocalDetuning object from lists of values and times for
        the magnitude.

        """
        magnitude = self.decode_physical_field(*magnitude_info)
        return ir.LocalDetuning(magnitude=magnitude)

    def decode_register(self, register: dict) -> ir.AtomArrangement:
        """Convert a dictionary to an AtomArrangement object."""
        return ir.AtomArrangement(
            sites=[[Decimal(value) for value in sublist] for sublist in register["sites"]],
            filling=register["filling"],
        )

    def decode_hamiltonian(self, data: dict) -> ir.Hamiltonian:
        """Convert a dictionary to a Hamiltonian object."""
        driving_fields = [
            self.decode_driving_field(
                (
                    field["amplitude"]["time_series"]["values"],
                    field["amplitude"]["time_series"]["times"],
                    field["amplitude"]["pattern"],
                ),
                (
                    field["phase"]["time_series"]["values"],
                    field["phase"]["time_series"]["times"],
                    field["phase"]["pattern"],
                ),
                (
                    field["detuning"]["time_series"]["values"],
                    field["detuning"]["time_series"]["times"],
                    field["detuning"]["pattern"],
                ),
            )
            for field in data["drivingFields"]
        ]

        local_detunings = [
            self.decode_local_detuning(
                (
                    term["magnitude"]["time_series"]["values"],
                    term["magnitude"]["time_series"]["times"],
                    term["magnitude"]["pattern"],
                )
            )
            for term in data["localDetuning"]
        ]

        return ir.Hamiltonian(drivingFields=driving_fields, localDetuning=local_detunings)

    def decode_ahs(self, data: dict) -> AnalogHamiltonianSimulation:
        """Convert a dictionary to an AnalogHamiltonianSimulation object."""
        register = self.decode_register(data["register"])
        hamiltonian = self.decode_hamiltonian(data["hamiltonian"])
        program_ir = ir.Program(setup=ir.Setup(ahs_register=register), hamiltonian=hamiltonian)
        return AnalogHamiltonianSimulation.from_ir(program_ir)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for ipython and other display functions.

"""
import logging
import sys
from unittest.mock import MagicMock, Mock, patch

import pytest

from qbraid._display import running_in_jupyter
from qbraid.programs.exceptions import PackageValueError
from qbraid.runtime._display import _job_table_jupyter
from qbraid.runtime.enums import JobStatus
from qbraid.runtime.native.provider import QbraidProvider

# pylint: disable=missing-function-docstring,redefined-outer-name

MOCK_JOB_DATA = [
    {
        "shots": 100,
        "circuitNumQubits": 1,
        "qbraidDeviceId": "aws_dm_sim",
        "vendorDeviceId": "arn:aws:braket:::device/quantum-simulator/amazon/dm1",
        "qbraidJobId": "aws_dm_sim-jovyan-qjob-1234567890",
        "status": "COMPLETED",
        "qbraidStatus": "COMPLETED",
        "vendor": "aws",
        "provider": "aws",
        "tags": {"test": "123"},
        "createdAt": "2024-04-30T17:28:38.962Z",
        "updatedAt": "2024-04-30T17:28:38.962Z",
    },
    {
        "shots": 100,
        "circuitNumQubits": 3,
        "qbraidDeviceId": "ibm_q_oslo",
        "vendorDeviceId": "ibm_oslo",
        "qbraidJobId": "ibm_q_oslo-jovyan-qjob-1234567890",
        "status": "COMPLETED",
        "qbraidStatus": "COMPLETED",
        "vendor": "ibm",
        "provider": "ibm",
        "tags": {"test": "*"},
        "createdAt": "2024-04-30T17:28:48.502Z",
        "updatedAt": "2024-04-30T17:28:48.502Z",
    },
]


class MockQbraidClient:
    """Mock QbraidClient class for testing"""

    def search_jobs(self, query: dict):
        """Mock search_jobs method."""
        max_results = query.get("resultsPerPage", 10)
        tag = query.get("tags.test", None)
        device_id = query.get("qbraidDeviceId", None)
        provider = query.get("provider", None)

        jobs = MOCK_JOB_DATA

        if device_id:
            jobs = [job for job in jobs if job["qbraidDeviceId"] == device_id]

        if provider:
            jobs = [job for job in jobs if job["provider"] == provider]

        if tag:
            tags = {"test": tag}

            jobs = [
                job
                for job in jobs
                if any(job["tags"].get(key) == value for key, value in tags.items())
            ]

        if max_results < len(jobs):
            return jobs[:max_results]
        return jobs


@pytest.fixture
def mock_provider():
    """Mock QbraidProvider class for testing."""
    return QbraidProvider(client=MockQbraidClient())


@pytest.fixture
def job_data():
    """Fixture to provide preprocessed job data."""
    JOB_STATUS_LIST = [status.name for status in JobStatus]
    data = [
        (f"job_{index}", f"timestamp_{index}", value) for index, value in enumerate(JOB_STATUS_LIST)
    ]
    yield data


class MockIPython:
    """Mock IPython class for testing."""

    def __init__(self, kernel):
        self.kernel = kernel


@pytest.fixture
def mock_ipython(request):
    """Fixture to mock IPython in sys.modules and restore it after the test."""
    kernel_value = request.param
    original_ipython = sys.modules.get("IPython", None)  # Save the original IPython module
    # Create a mock module and assign get_ipython
    mock_module = type("MockModule", (), {})()  # Create a mock module object
    mock_module.get_ipython = Mock(return_value=MockIPython(kernel_value))
    sys.modules["IPython"] = mock_module  # Set the mock module in sys.modules

    yield  # Yield control to the test function

    # Restore the original state after the test
    if original_ipython is not None:
        sys.modules["IPython"] = original_ipython
    else:
        del sys.modules["IPython"]


def test_package_value_error():
    """Test raising PackageValueError exception."""
    with pytest.raises(PackageValueError):
        raise PackageValueError("custom msg")


def test_running_in_jupyter():
    """Test ``running_in_jupyter`` for non-jupyter environment."""
    assert not running_in_jupyter()


@pytest.mark.parametrize("mock_ipython", [None], indirect=True)
def test_ipython_imported_but_ipython_none(mock_ipython):
    """Test ``running_in_jupyter`` for IPython imported but ``get_ipython()`` returns None."""
    assert not running_in_jupyter()


@pytest.mark.parametrize("mock_ipython", ["non-empty kernel"], indirect=True)
def test_ipython_imported_and_in_jupyter(mock_ipython):
    """Test ``running_in_jupyter`` for IPython imported and in Jupyter."""
    assert running_in_jupyter()


@pytest.mark.parametrize("mock_ipython", [None], indirect=True)
def test_get_jobs_no_results(capfd, mock_provider, mock_ipython):
    """Test ``get_jobs`` stdout for results == 0.
    When no results are found, a single line is printed.
    """
    mock_provider.display_jobs(device_id="non-existent-device", status="ONLINE")
    out, err = capfd.readouterr()
    assert out == "No jobs found matching given criteria\n"
    assert len(err) == 0


@pytest.mark.parametrize("mock_ipython", [None], indirect=True)
def test_get_aws_jobs_by_tag(capfd, mock_provider, mock_ipython):
    """Test ``get_jobs`` for aws tagged jobs."""
    mock_provider.display_jobs(tags={"test": "123"}, provider="aws", max_results=1)
    out, err = capfd.readouterr()
    message = out.split("\n")[0]
    assert message == "Displaying 1 most recent job matching query:"
    assert MOCK_JOB_DATA[0]["qbraidJobId"] in out
    assert len(err) == 0


@pytest.mark.parametrize("mock_ipython", [None], indirect=True)
def test_get_ibm_jobs_by_tag(capfd, mock_provider, mock_ipython):
    """Test ``get_jobs`` for ibm tagged jobs."""
    mock_provider.display_jobs(tags={"test": "*"}, provider="ibm", max_results=1)
    out, err = capfd.readouterr()
    message = out.split("\n")[0]
    assert message == "Displaying 1 most recent job matching query:"
    assert MOCK_JOB_DATA[1]["qbraidJobId"] in out
    assert len(err) == 0


@pytest.mark.parametrize("mock_ipython", [None], indirect=True)
def test_get_jobs_results(capfd, mock_provider, mock_ipython):
    """Test ``get_jobs`` stdout for results > 0.
    When results returned, output format is as follows:
    (1) Message
    (2) Empty line
    (3) Section titles
    (4) Underline titles
    (4+x) ``x`` lines of results
    (5+x) Empty line

    So, for ``numResults == x`` we expected ``6+x`` total lines from stdout.
    """
    num_results = 2  # test value
    lines_expected = 5 + num_results
    mock_provider.display_jobs(max_results=num_results)
    out, err = capfd.readouterr()
    lines_out = len(out.split("\n"))
    assert lines_out == lines_expected
    assert len(err) == 0


@pytest.mark.parametrize("mock_ipython", ["non-empty kernel"], indirect=True)
def test_display_jobs_in_jupyter(capfd, job_data, mock_ipython):
    """Test ``_display_jobs_jupyter`` stdout for non-empty job status list."""
    msg = "test123"
    _job_table_jupyter(job_data, msg)
    out, err = capfd.readouterr()
    assert "IPython.core.display.HTML object" in out
    assert len(err) == 0


@pytest.mark.parametrize("mock_ipython", ["non-empty kernel"], indirect=True)
def test_get_jobs_in_jupyter(capfd, mock_provider, mock_ipython):
    """Test `get_jobs` stdout for non-empty kernel.
    When running in Jupyter, the output should be an HTML object."""
    mock_provider.display_jobs()  # Assuming this uses IPython.get_ipython internally
    out, err = capfd.readouterr()
    assert "IPython.core.display.HTML object" in out
    assert len(err) == 0


def test_running_in_jupyter_logs_exception(caplog):
    """
    Test that an exception raised in running_in_jupyter() is logged correctly.
    """
    with patch("sys.modules", new_callable=lambda: {"IPython": MagicMock()}):
        sys.modules["IPython"].__dict__["get_ipython"] = MagicMock(
            side_effect=Exception("Test exception")
        )

        caplog.set_level(logging.ERROR)

        result = running_in_jupyter()

        assert not result, "running_in_jupyter() should return False when an exception occurs"
        assert (
            "Error checking if running in Jupyter: Test exception" in caplog.text
        ), "The exception should be logged as an error"



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing Cirq conversion extras.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import requires_extras

stimcirq = LazyLoader("stimcirq", globals(), "stimcirq")
qbraid_qir = LazyLoader("qbraid_qir", globals(), "qbraid_qir")

if TYPE_CHECKING:
    import cirq
    import pyqir  # type: ignore
    import stim  # type: ignore


@requires_extras("stim", "stimcirq")
def cirq_to_stim(circuit: cirq.Circuit) -> stim.Circuit:
    """Returns an stim circuit equivalent to the input cirq circuit.

    Args:
        circuit (cirq.Circuit): cirq circuit to convert to stim circuit.

    Returns:
        stim.Circuit: stim circuit equivalent to input cirq circuit.
    """
    return stimcirq.cirq_circuit_to_stim_circuit(circuit)


@requires_extras("stim", "stimcirq")
def stim_to_cirq(circuit: stim.Circuit) -> cirq.Circuit:
    """Returns a Cirq circuit equivalent to the input Stim circuit.

    Args:
        circuit (stim.Circuit): Stim circuit to convert to Cirq circuit.

    Returns:
        cirq.Circuit: Cirq circuit equivalent to input Stim circuit.
    """
    return stimcirq.stim_circuit_to_cirq_circuit(circuit)


@requires_extras("qbraid_qir")
def cirq_to_pyqir(circuit: cirq.Circuit) -> pyqir.Module:
    """Returns a PyQIR module equivalent to the input cirq circuit.

    Args:
        circuit (cirq.Circuit): cirq circuit to convert to PyQIR module.

    Returns:
        pyqir.Module: module equivalent to input cirq circuit.
    """
    return qbraid_qir.cirq.cirq_to_qir(circuit)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Qiskit OpenQASM conversions

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qiskit.qasm2 import dumps as qasm2_dumps

from qbraid.transpiler.annotations import weight

if TYPE_CHECKING:
    import qiskit as qiskit_

    from qbraid.programs.typer import Qasm2StringType


@weight(0.999)
def qiskit_to_qasm2(circuit: qiskit_.QuantumCircuit) -> Qasm2StringType:
    """Returns OpenQASM 2 string equivalent to the input Qiskit circuit.

    Args:
        circuit: Qiskit circuit to convert to OpenQASM 2 string.

    Returns:
        str: OpenQASM 2 representation of the input Qiskit circuit.
    """
    return qasm2_dumps(circuit)



# Copyright (C) 2024 qBraid
# Copyright (C) Unitary Fund
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# This file includes code adapted from Mitiq (https://github.com/unitaryfund/mitiq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module containing function to test Cirq circuit equality

"""
from copy import deepcopy

import cirq

import qbraid.programs.gate_model.cirq


def _rev_qubits(circuit: cirq.Circuit) -> cirq.Circuit:
    """Reverses the qubit order of a circuit."""
    program = qbraid.programs.gate_model.cirq.CirqCircuit(circuit)
    program.reverse_qubit_order()
    return program.program


def remove_empty_moments(circuit: cirq.Circuit) -> cirq.Circuit:
    """Remove all empty moments from a Cirq circuit.

    Args:
        circuit (cirq.Circuit): The input quantum circuit to clean up.

    Returns:
        cirq.C ircuit: A new circuit with all empty moments removed.
    """
    non_empty_moments = [moment for moment in circuit if moment.operations]
    return cirq.Circuit(non_empty_moments)


def compare_circuits(circuit1: cirq.Circuit, circuit2: cirq.Circuit):
    """Compares two Cirq circuits and prints differences.

    Args:
        circuit1 (cirq.Circuit): The first circuit to compare.
        circuit2 (cirq.Circuit): The second circuit to compare.
    """
    # First, check if they have the same number of moments
    if len(circuit1) != len(circuit2):
        print(f"Circuits differ in number of moments: {len(circuit1)} vs {len(circuit2)}")
        return

    # Iterate through each moment
    for i, (moment1, moment2) in enumerate(zip(circuit1, circuit2)):
        if moment1 != moment2:
            print(f"Difference found in moment {i}")
            ops1 = set(moment1.operations)
            ops2 = set(moment2.operations)

            # Compare operations in each moment
            if ops1 != ops2:
                unique_to_circuit1 = ops1 - ops2
                unique_to_circuit2 = ops2 - ops1
                if unique_to_circuit1:
                    print(f" Unique to circuit1 in moment {i}: {unique_to_circuit1}")
                if unique_to_circuit2:
                    print(f" Unique to circuit2 in moment {i}: {unique_to_circuit2}")


def _equal(
    circuit_one: cirq.Circuit,
    circuit_two: cirq.Circuit,
    require_qubit_equality: bool = False,
    require_measurement_equality: bool = False,
    allow_reversed_qubit_order: bool = False,
) -> bool:
    """Returns True if the circuits are equal, else False.

    Args:
        circuit_one: Input circuit to compare to circuit_two.
        circuit_two: Input circuit to compare to circuit_one.
        require_qubit_equality: Requires that the qubits be equal
            in the two circuits.
        require_measurement_equality: Requires that measurements are equal on
            the two circuits, meaning that measurement keys are equal.
        allow_reversed_qubit_order: Allows considering the circuits as equal
            even if their qubit order is reversed. Default is False.

    Note:
        If set(circuit_one.all_qubits()) = {LineQubit(0)},
        then set(circuit_two_all_qubits()) must be {LineQubit(0)},
        else the two are not equal.
        If True, the qubits of both circuits must have a well-defined ordering.
    """
    # Make a deepcopy only if it's necessary
    if not (require_qubit_equality and require_measurement_equality):
        circuit_one = deepcopy(circuit_one)
        circuit_two = deepcopy(circuit_two)

    if not require_qubit_equality:
        # Transform the qubits of circuit one to those of circuit two
        qubit_map = dict(
            zip(
                sorted(circuit_one.all_qubits()),
                sorted(circuit_two.all_qubits()),
            )
        )
        circuit_one = circuit_one.transform_qubits(lambda q: qubit_map[q])

    if not require_measurement_equality:
        for circ in (circuit_one, circuit_two):
            measurements = [
                (moment, op)
                for moment, op, _ in circ.findall_operations_with_gate_type(cirq.MeasurementGate)
            ]
            circ.batch_remove(measurements)

    if allow_reversed_qubit_order and circuit_one != circuit_two:
        reversed_circuit_one = _rev_qubits(circuit_one)
        return _equal(
            reversed_circuit_one,
            circuit_two,
            require_qubit_equality,
            require_measurement_equality,
            False,
        )

    circuit_one_clean = remove_empty_moments(circuit_one)
    circuit_two_clean = remove_empty_moments(circuit_two)

    return circuit_one_clean == circuit_two_clean



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for the qbraid visualization plot counts functions

"""

from unittest.mock import patch

import pytest

from qbraid.visualization.plot_counts import _counts_to_decimal, plot_distribution, plot_histogram

PLOT_FNS = [plot_histogram, plot_distribution]


def test_counts_to_decimal_normal_case():
    """Test convering measurement counts to decimal probabilities."""
    counts_dict = {"00": 10, "01": 15, "10": 20, "11": 5}
    expected_output = {"00": 0.2, "01": 0.3, "10": 0.4, "11": 0.1}
    assert _counts_to_decimal(counts_dict) == expected_output


def test_counts_to_decimal_with_zero_total_count():
    """Test raising error when total of counts is zero."""
    counts_dict = {"00": 0, "01": 0, "10": 0, "11": 0}
    with pytest.raises(ValueError, match="Total count cannot be zero."):
        _counts_to_decimal(counts_dict)


def test_counts_to_decimal_with_non_integer_values():
    """Test raising error when counts values are not integers."""
    counts_dict = {"00": "ten", "01": 15, "10": 20, "11": 5}
    with pytest.raises(TypeError, match="Counts values must be integers."):
        _counts_to_decimal(counts_dict)


@pytest.mark.parametrize("plot_function", PLOT_FNS)
def test_plot_counts_single_dict(plot_function):
    """Test plotting histogram with single counts dict."""
    with (
        patch("matplotlib.pyplot.show") as mock_show,
        patch("matplotlib.pyplot.savefig") as mock_save,
    ):
        counts_dict = {"00": 50, "01": 30, "10": 10, "11": 10}
        plot_function(
            counts_dict,
            legend="Dict",
            colors="crimson",
            x_label="States",
            title="Single Dict Test",
            save_path="fake/path",
            show_plot=True,
        )
        mock_show.assert_called_once()
        mock_save.assert_called_once()

        plot_function(counts_dict)


@pytest.mark.parametrize("plot_function", PLOT_FNS)
def test_plot_histogram_multiple_dicts(plot_function):
    """Test plotting histogram with multiple counts dicts."""
    counts_dict1 = {"00": 50, "01": 30, "10": 10, "11": 10}
    counts_dict2 = {"00": 20, "01": 40, "10": 30, "11": 10}

    plot_function(
        counts=[counts_dict1, counts_dict2],
        legend=["First Execution", "Second Execution"],
        colors=["crimson", "midnightblue"],
        title="Multiple Dicts Test",
        show_plot=False,
    )


@pytest.mark.parametrize("plot_function", PLOT_FNS)
def test_plot_histogram_mismatched_legend_length(plot_function):
    """Test raising error when legend length does not match counts length."""
    counts_dict1 = {"00": 50, "01": 30, "10": 10, "11": 10}
    counts_dict2 = {"00": 20, "01": 40, "10": 30, "11": 10}

    with pytest.raises(ValueError):
        plot_function(
            counts=[counts_dict1, counts_dict2],
            legend=["Only one label"],
            colors=["crimson", "midnightblue"],
            show_plot=False,
        )


@pytest.mark.parametrize("plot_function", PLOT_FNS)
def test_plot_histogram_mismatched_colors_length(plot_function):
    """Test raising error when colors length does not match counts length."""
    counts_dict1 = {"00": 50, "01": 30, "10": 10, "11": 10}
    counts_dict2 = {"00": 20, "01": 40, "10": 30, "11": 10}

    with pytest.raises(ValueError):
        plot_function(
            counts=[counts_dict1, counts_dict2],
            legend=["First Execution", "Second Execution"],
            colors=["only one color"],
            show_plot=False,
        )



# Copyright (C) 2024 qBraid
# Copyright (C) Unitary Fund
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# This file includes code adapted from Mitiq (https://github.com/unitaryfund/mitiq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Unit tests for converting Braket circuits to Cirq circuits

"""
import cirq
import numpy as np
import pytest
from braket.circuits import Circuit as BKCircuit
from braket.circuits import Gate as BKGate
from braket.circuits import Instruction, QubitSet
from braket.circuits import gates as braket_gates
from braket.circuits import noises as braket_noise_gate
from braket.circuits.serialization import OpenQASMSerializationProperties
from braket.ir.jaqcd.instructions import Unitary as JaqcdUnitary
from braket.pulse import Frame, Port
from braket.pulse.pulse_sequence import PulseSequence
from cirq import LineQubit
from cirq import ops as cirq_ops

from qbraid.interface import circuits_allclose
from qbraid.transpiler.conversions.braket import braket_to_cirq
from qbraid.transpiler.conversions.braket.braket_to_cirq import (
    _from_braket_instruction,
    _from_one_qubit_braket_instruction,
    _from_two_qubit_braket_instruction,
    _give_cirq_gate_name,
)
from qbraid.transpiler.conversions.cirq.braket_custom import C as BKControl
from qbraid.transpiler.exceptions import ProgramConversionError


def test_from_braket_bell_circuit():
    """Test converting bell circuit"""
    braket_circuit = BKCircuit().h(0).cnot(0, 1)  # pylint: disable=no-member
    cirq_circuit = braket_to_cirq(braket_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


def test_from_braket_non_parameterized_single_qubit_gates():
    """Test converting circuit containing non-parameterized single-qubits gates"""
    braket_circuit = BKCircuit()
    instructions = [
        Instruction(braket_gates.I(), target=0),
        Instruction(braket_gates.X(), target=1),
        Instruction(braket_gates.Y(), target=2),
        Instruction(braket_gates.Z(), target=3),
        Instruction(braket_gates.H(), target=0),
        Instruction(braket_gates.S(), target=1),
        Instruction(braket_gates.Si(), target=2),
        Instruction(braket_gates.T(), target=3),
        Instruction(braket_gates.Ti(), target=0),
        Instruction(braket_gates.V(), target=1),
        Instruction(braket_gates.Vi(), target=2),
    ]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = braket_to_cirq(braket_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubit_index", (0, 3))
def test_from_braket_parameterized_single_qubit_gates(qubit_index):
    """Testing converting circuit containing parameterized single-qubit gates"""
    braket_circuit = BKCircuit()
    pgates = [
        braket_gates.Rx,
        braket_gates.Ry,
        braket_gates.Rz,
        braket_gates.PhaseShift,
    ]
    angles = np.random.RandomState(11).random(len(pgates))  # pylint: disable=no-member
    instructions = [Instruction(rot(a), target=qubit_index) for rot, a in zip(pgates, angles)]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = braket_to_cirq(braket_circuit)

    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


def test_from_braket_non_parameterized_two_qubit_gates():
    """Test converting circuit containing non-parameterized two-qubit gates."""
    braket_circuit = BKCircuit()
    instructions = [
        Instruction(braket_gates.CNot(), target=[2, 3]),
        Instruction(braket_gates.Swap(), target=[3, 4]),
        Instruction(braket_gates.ISwap(), target=[2, 3]),
        Instruction(braket_gates.CZ(), target=(3, 4)),
        Instruction(braket_gates.CY(), target=(2, 3)),
    ]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = braket_to_cirq(braket_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


def test_from_braket_parameterized_two_qubit_gates():
    """Test converting circuit containing parameterized two-qubit gates."""
    braket_circuit = BKCircuit()
    pgates = [
        braket_gates.CPhaseShift,
        braket_gates.CPhaseShift00,
        braket_gates.CPhaseShift01,
        braket_gates.CPhaseShift10,
        braket_gates.PSwap,
        braket_gates.XX,
        braket_gates.YY,
        braket_gates.ZZ,
        braket_gates.XY,
    ]
    angles = np.random.RandomState(2).random(len(pgates))  # pylint: disable=no-member
    instructions = [Instruction(rot(a), target=[0, 1]) for rot, a in zip(pgates, angles)]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = braket_to_cirq(braket_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


def test_from_braket_three_qubit_gates():
    """Testing converting circuits containing three-qubit gates"""
    braket_circuit = BKCircuit()
    instructions = [
        Instruction(braket_gates.CCNot(), target=[1, 2, 3]),
        Instruction(braket_gates.CSwap(), target=[1, 2, 3]),
    ]
    for instr in instructions:
        braket_circuit.add_instruction(instr)
    cirq_circuit = braket_to_cirq(braket_circuit)
    assert circuits_allclose(braket_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize(
    "noise_gate, target_gate",
    [
        (braket_noise_gate.BitFlip, cirq_ops.BitFlipChannel),
        (braket_noise_gate.PhaseFlip, cirq_ops.PhaseFlipChannel),
        (braket_noise_gate.Depolarizing, cirq_ops.DepolarizingChannel),
    ],
)
def test_single_probability_noise_gate(noise_gate, target_gate):
    """Testing converting circuits containing one-probability noise gates"""
    braket_circuit = BKCircuit()
    probs = np.random.uniform(low=0, high=0.5)  # pylint: disable=no-member
    instructions = Instruction(noise_gate(probs), target=[0])
    braket_circuit.add_instruction(instructions)
    cirq_circuit = braket_to_cirq(braket_circuit)
    gate = list(cirq_circuit.all_operations())[0].gate
    assert isinstance(gate, target_gate)
    assert gate.p == probs


@pytest.mark.parametrize(
    "noise_gate, target_gate",
    [
        (braket_noise_gate.AmplitudeDamping, cirq_ops.AmplitudeDampingChannel),
        (braket_noise_gate.PhaseDamping, cirq_ops.PhaseDampingChannel),
    ],
)
def test_single_gamma_noise_gate(noise_gate, target_gate):
    """Testing converting circuits containing one-probability noise gates"""
    braket_circuit = BKCircuit()
    probs = np.random.uniform(low=0, high=0.5)  # pylint: disable=no-member
    instructions = Instruction(noise_gate(probs), target=[0])
    braket_circuit.add_instruction(instructions)
    cirq_circuit = braket_to_cirq(braket_circuit)
    gate = list(cirq_circuit.all_operations())[0].gate
    assert isinstance(gate, target_gate)
    assert gate.gamma == probs


def test_kraus_gates():
    """Testing converting Kraus noise gates"""
    K0 = np.sqrt(0.8) * np.eye(4)
    K1 = np.sqrt(0.2) * np.kron(np.array([[0, 1], [1, 0]]), np.array([[0, 1], [1, 0]]))
    instructions = Instruction(braket_noise_gate.Kraus(matrices=[K0, K1]), target=[0, 1])
    braket_circuit = BKCircuit().add_instruction(instructions)
    cirq_circuit = braket_to_cirq(braket_circuit)
    gate = list(cirq_circuit.all_operations())[0].gate
    assert isinstance(gate, cirq_ops.kraus_channel.KrausChannel)
    assert np.allclose(gate._kraus_ops, [K0, K1])


def test_generalized_amplitude_damping_channel_gate():
    """Testing converting Kraus noise gates"""
    probs = np.random.uniform(low=0, high=0.5, size=2)
    instruction = Instruction(
        braket_noise_gate.GeneralizedAmplitudeDamping(gamma=probs[0], probability=probs[1]),
        target=[0],
    )
    braket_circuit = BKCircuit().add_instruction(instruction)
    cirq_circuit = braket_to_cirq(braket_circuit)
    gate = list(cirq_circuit.all_operations())[0].gate
    assert isinstance(gate, cirq_ops.GeneralizedAmplitudeDampingChannel)
    assert gate.gamma == probs[0]
    assert gate.p == probs[1]


def test_depolarizing_channel_gate():
    """Testing converting Kraus noise gates"""
    probs = np.random.uniform(low=0, high=0.5, size=1)
    instruction = Instruction(
        braket_noise_gate.TwoQubitDepolarizing(probability=probs[0]), target=[0, 1]
    )
    braket_circuit = BKCircuit().add_instruction(instruction)
    cirq_circuit = braket_to_cirq(braket_circuit)
    gate = list(cirq_circuit.all_operations())[0].gate
    assert isinstance(gate, cirq_ops.common_channels.DepolarizingChannel)
    assert gate.p == probs
    assert gate.n_qubits == 2


def test_convert_ionq_gates():
    """Test converting IonQ GPi, GPi2, and MS (Mølmer-Sørenson) gates."""
    bk_circuit = BKCircuit()
    bk_circuit.gpi(0, np.pi)
    bk_circuit.gpi2(1, np.pi / 3)
    bk_circuit.ms(0, 1, np.pi / 4, np.pi / 2, 3 * np.pi / 4)
    cirq_circuit = braket_to_cirq(bk_circuit)
    assert circuits_allclose(bk_circuit, cirq_circuit, strict_gphase=True)


def test_raise_error():
    """Test raising error when converting unsupported Pulse sequences"""
    with pytest.raises(ProgramConversionError):
        pre_fram = Frame(
            frame_id="predefined_frame_1",
            frequency=2e9,
            port=Port(port_id="device_port_x0", dt=1e-9, properties={}),
            phase=0,
            is_predefined=True,
        )
        pulse_seq = PulseSequence().set_frequency(pre_fram, 6e6)

        test_case = BKCircuit().add_instruction(
            Instruction(braket_gates.PulseGate(pulse_seq, 1), [0])
        )
        braket_to_cirq(test_case)


def _rotation_of_pi_over_7(num_qubits):
    matrix = np.identity(2**num_qubits)
    matrix[0:2, 0:2] = [
        [np.cos(np.pi / 7), np.sin(np.pi / 7)],
        [-np.sin(np.pi / 7), np.cos(np.pi / 7)],
    ]
    return matrix


def test_from_braket_raises_on_unsupported_gates():
    """Test that converting circuit with unsupported gate raises error"""
    for num_qubits in range(1, 5):
        braket_unitary_circuit = BKCircuit()
        instr = Instruction(
            braket_gates.Unitary(_rotation_of_pi_over_7(num_qubits)),
            target=list(range(num_qubits)),
        )
        braket_unitary_circuit.add_instruction(instr)
    with pytest.raises(ProgramConversionError):
        braket_to_cirq(braket_unitary_circuit)


def test_braket_control_custom():
    """Test converting Braket controlled gate with custom control"""
    sub_gate = braket_gates.X()
    targets = QubitSet([0, 1])
    props = OpenQASMSerializationProperties()
    control_gate = BKControl(sub_gate, targets)
    qasm = control_gate._to_openqasm(target=targets, serialization_properties=props)
    expected = (
        "#pragma braket unitary([[1.0, 0, 0, 0], [0, 1.0, 0, 0], "
        "[0, 0, 0, 1.0], [0, 0, 1.0, 0]]) q[0], q[1]"
    )
    assert qasm == expected
    adjoint_gates = control_gate.adjoint()
    assert isinstance(adjoint_gates, list)
    assert len(adjoint_gates) == 1
    adjoint = adjoint_gates[0]
    assert isinstance(adjoint, braket_gates.Unitary)
    assert adjoint.qubit_count == 2
    unitary = control_gate._to_jaqcd(target=targets)
    assert isinstance(unitary, JaqcdUnitary)


def test_cirq_circuit_diagram_info():
    """Test that circuit diagram info is correctly set"""
    id_gate = cirq_ops.IdentityGate(1)
    _give_cirq_gate_name(id_gate, "I", 1)
    assert id_gate._circuit_diagram_info_(None) == ("I",)


def test_from_braket_instruction_measure():
    """Test from_braket_instruction with Measure gate"""
    circuit = BKCircuit().measure(0)
    instr = circuit.instructions[0]
    bk_qubits = [int(q) for q in circuit.qubits]
    cirq_qubits = [cirq.LineQubit(x) for x in bk_qubits]
    qubit_mapping = {q: cirq_qubits[i] for i, q in enumerate(bk_qubits)}
    cirq_instr = _from_braket_instruction(instr, qubit_mapping)
    assert isinstance(cirq_instr[0]._gate, cirq.MeasurementGate)


def test_from_braket_instruction_one_qubit():
    """Test from_braket_instruction with one-qubit gate"""
    gate = BKGate(1, "1")
    instr = Instruction(gate, [0])
    qubit_mapping = {0: LineQubit(0)}
    instr.operator.to_matrix = lambda: np.eye(3)
    with pytest.raises(ValueError):
        _from_one_qubit_braket_instruction(instr, qubit_mapping)


def test_from_braket_instruction_two_qubit():
    """Test from_braket_instruction with one-qubit gate"""
    gate = BKGate(2, "12")
    instr = Instruction(gate, [0, 1])
    qubit_mapping = {0: LineQubit(0), 1: LineQubit(1)}
    instr.operator.to_matrix = lambda: np.eye(3)
    with pytest.raises(ValueError):
        _from_two_qubit_braket_instruction(instr, qubit_mapping)


def test_from_braket_instruction_three_qubit():
    """Test from_braket_instruction with three-qubit gate"""
    gate = BKGate(3, "123")
    instr = Instruction(gate, [0, 1, 2])
    qubit_mapping = {0: LineQubit(0), 1: LineQubit(1), 2: LineQubit(2)}
    instr.operator.to_matrix = lambda: np.eye(7)
    with pytest.raises(ProgramConversionError):
        _from_braket_instruction(instr, qubit_mapping)

    instr.operator.to_matrix = lambda: np.eye(8)
    cirq_instr = _from_braket_instruction(instr, qubit_mapping)
    assert isinstance(cirq_instr[0].gate, cirq.MatrixGate)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint:disable=invalid-name

"""
Module defining abstract QuantumDevice Class

"""
from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from qbraid._logging import logger
from qbraid.programs import (
    ProgramLoaderError,
    ProgramSpec,
    ProgramTypeError,
    get_program_type_alias,
    load_program,
)
from qbraid.transpiler import (
    ConversionGraph,
    ConversionPathNotFoundError,
    ConversionScheme,
    ProgramConversionError,
    transpile,
)

from .enums import DeviceStatus, ValidationLevel
from .exceptions import ProgramValidationError, ResourceNotFoundError
from .options import RuntimeOptions

if TYPE_CHECKING:
    import qbraid.programs
    import qbraid.runtime
    import qbraid.transpiler


class QuantumDevice(ABC):
    """Abstract interface for quantum devices."""

    def __init__(
        self,
        profile: qbraid.runtime.TargetProfile,
        scheme: Optional[ConversionScheme] = None,
        options: Optional[RuntimeOptions] = None,
    ):
        """Create a ``QuantumDevice`` object.

        Args:
            profile (TargetProfile): The device runtime profile.
            scheme (Optional[ConversionScheme]): The conversion graph and options passed
                to the transpiler at runtime.
            options (Optional[RuntimeOptions]): Custom options to control the runtime
                behavior. Adds fields or overrides default values for ``transpile``,
                ``transform``, and ``validate``. Note that while you can modify these
                values, their associated validators are fixed and cannot be changed.
        """
        self._profile = profile
        self._target_spec: Optional[Union[ProgramSpec, list[ProgramSpec]]] = profile.program_spec
        self._scheme = scheme or ConversionScheme()
        self._options = self._default_options()
        if options:
            self._options.merge(options, override_validators=False)

    @property
    def profile(self) -> qbraid.runtime.TargetProfile:
        """Return the runtime profile."""
        return self._profile

    @property
    def id(self) -> str:
        """Return the device ID."""
        return self.profile.device_id

    @property
    def num_qubits(self) -> Optional[int]:
        """The number of qubits supported by the device."""
        return self.profile.num_qubits

    @property
    def simulator(self) -> bool:
        """The device type, Simulator, Fake_device or QPU."""
        return self.profile.simulator

    @property
    def scheme(self) -> ConversionScheme:
        """Return the conversion scheme."""
        if not self._scheme.conversion_graph:
            self._scheme.update_values(conversion_graph=ConversionGraph(include_isolated=True))
        return self._scheme

    def __repr__(self):
        """Return a string representation of the device."""
        return f"<{self.__module__}.{self.__class__.__name__}('{self.id}')>"

    @abstractmethod
    def status(self) -> qbraid.runtime.DeviceStatus:
        """Return device status."""

    @classmethod
    def _default_options(cls) -> RuntimeOptions:
        """Define default options for the QuantumDevice."""
        options = RuntimeOptions(
            transpile=True, transform=True, validate=ValidationLevel.RAISE, prepare=True
        )

        # pylint: disable=unnecessary-lambda
        options.set_validator("transpile", lambda x: isinstance(x, bool))
        options.set_validator("transform", lambda x: isinstance(x, bool))
        options.set_validator(
            "validate",
            lambda x: isinstance(x, ValidationLevel) or (isinstance(x, int) and 0 <= x <= 2),
        )
        options.set_validator("prepare", lambda x: isinstance(x, bool))

        # pylint: enable=unnecessary-lambda

        return options

    def set_options(self, **fields):
        """
        Update the runtime options for the QuantumDevice.

        The runtime options control the default behavior of the `QuantumDevice.run` method,
        including settings such as transpilation, verification, and transformation. If an
        unsupported option is provided, an `AttributeError` will be raised.

        Args:
            **fields: Keyword arguments representing the runtime options to update.
                    The options must already exist in the device's configuration.

        Raises:
            AttributeError: If an invalid runtime option is passed.
        """
        for field, value in fields.items():
            if not hasattr(self._options, field):
                raise AttributeError(f"Options field '{field}' is not valid for this device")
            if field == "validate" and isinstance(
                value, bool
            ):  # TODO: Move this to the RuntimeOptions class
                fields[field] = ValidationLevel.RAISE if value else ValidationLevel.NONE
        self._options.update_options(**fields)

    def queue_depth(self) -> int:
        """Return the number of jobs in the queue for the device."""
        raise ResourceNotFoundError("Queue depth is not available for this device.")

    def avg_queue_time(self) -> int:
        """Return the average time (in seconds) a job spends in the queue for the device."""
        raise ResourceNotFoundError("Average queue time is not available for this device.")

    def update_scheme(self, **kwargs):
        """Update the conversion scheme with new values."""
        self._scheme.update_values(**kwargs)

    def metadata(self) -> dict[str, Any]:
        """
        Returns a dictionary containing selected metadata about the device.

        The metadata excludes the program specifications, and it includes
        the device's current status and queue depth.

        Returns:
            dict[str, Any]: A dictionary with device status and queue depth among other details.
        """
        metadata = self.profile.model_dump(
            exclude=["program_spec", "experiment_type", "noise_models"]
        )

        try:
            metadata["queue_depth"] = self.queue_depth()
        except ResourceNotFoundError as err:
            logger.info(err)

        try:
            metadata["average_queue_time"] = self.avg_queue_time()
        except ResourceNotFoundError as err:
            logger.info(err)

        metadata["status"] = self.status().name
        metadata["paradigm"] = (
            self.profile.experiment_type.value if self.profile.experiment_type else None
        )

        if self.simulator is True:
            metadata["noise_models"] = (
                list(self.profile.noise_models) if self.profile.noise_models else None
            )

        options = {
            key: (value.value if isinstance(value, ValidationLevel) else value)
            for key, value in dict(self._options).items()
        }

        program_spec = self._target_spec

        if not program_spec:
            target_program_type = None
        elif isinstance(program_spec, list):
            target_program_type = [ps.alias for ps in program_spec]
        else:
            target_program_type = program_spec.alias

        runtime_config = {
            "target_program_type": target_program_type,
            "conversion_scheme": self._scheme.to_dict(),
            "options": options,
        }

        metadata["runtime_config"] = runtime_config

        return metadata

    def set_target_program_type(self, alias: str | list[str] | None) -> None:
        """Set the program type to target during runtime transpile step.

        Args:
            alias: The alias(es) of the target program spec(s) to set.

        Raises:
            ValueError: If the given alias does not match any program spec in the target profile.
        """
        if alias is None:
            self._target_spec = None
            return

        aliases = [alias] if isinstance(alias, str) else alias

        if len(set(aliases)) != len(aliases):
            raise ValueError("Duplicate aliases are not allowed.")

        if isinstance(self.profile.program_spec, list):
            spec_aliases = {spec.alias for spec in self.profile.program_spec}
            if isinstance(alias, str):
                if alias not in spec_aliases:
                    raise ValueError(
                        f"Invalid alias: '{alias}'. Available aliases: {', '.join(spec_aliases)}"
                    )
                self._target_spec = next(
                    spec for spec in self.profile.program_spec if spec.alias == alias
                )
            else:
                missing_aliases = [a for a in aliases if a not in spec_aliases]
                if missing_aliases:
                    raise ValueError(
                        f"Invalid aliases: {', '.join(missing_aliases)}. "
                        f"Available aliases: {', '.join(spec_aliases)}"
                    )
                self._target_spec = [
                    spec for spec in self.profile.program_spec if spec.alias in aliases
                ]

        elif self.profile.program_spec is not None:
            if len(aliases) != 1:
                raise ValueError(
                    "Alias list must contain exactly one alias when the target profile "
                    "has a single program spec."
                )
            as_lst = isinstance(alias, list)
            alias = aliases[0]
            if alias != self.profile.program_spec.alias:
                raise ValueError(
                    f"Invalid alias: '{alias}'. "
                    f"Available aliases: '{self.profile.program_spec.alias}'"
                )
            self._target_spec = [self.profile.program_spec] if as_lst else self.profile.program_spec

        else:
            raise ValueError("Target profile has no program spec defined.")

    def _get_target_spec(self, run_input: qbraid.programs.QPROGRAM) -> ProgramSpec:
        run_input_alias = get_program_type_alias(run_input, safe=True)
        target_specs = (
            self._target_spec
            if isinstance(self._target_spec, list)
            else [self._target_spec] if self._target_spec else []
        )

        for target_spec in target_specs:
            if target_spec.alias == run_input_alias:
                return target_spec

        raise ProgramTypeError(
            message=f"Could not find a target ProgramSpec matching the alias '{run_input_alias}'."
        )

    def transpile(
        self,
        run_input: qbraid.programs.QPROGRAM,
        run_input_spec: qbraid.programs.ProgramSpec,
    ) -> qbraid.programs.QPROGRAM:
        """Convert circuit to package compatible with target device and pass through any
        provider transpile methods to match topology of device and/or optimize as applicable.

        Returns:
            :data:`~qbraid.programs.QPROGRAM`: Transpiled quantum program

        Raises:
            ProgramConversionError: If program conversion fails

        """
        if not self._target_spec:
            logger.info("Skipping transpile: no target ProgramSpec specified in TargetProfile.")
            return run_input

        graph = self.scheme.conversion_graph
        target_specs = (
            self._target_spec if isinstance(self._target_spec, list) else [self._target_spec]
        )

        alias_to_spec = {target_spec.alias: target_spec for target_spec in target_specs}
        ordered_targets = graph.get_sorted_closest_targets(
            run_input_spec.alias, list(alias_to_spec.keys())
        )
        ordered_target_specs = [alias_to_spec[alias] for alias in ordered_targets]

        cached_errors = []
        conversion_scheme_fields = self.scheme.to_dict()

        for target_spec in ordered_target_specs:
            target_alias = target_spec.alias
            target_type = target_spec.program_type

            if run_input_spec.alias == target_alias:
                return run_input

            logger.debug(
                "Transpiling '%s' program to device target program type '%s'",
                run_input_spec.alias,
                target_alias,
            )

            try:
                transpiled_run_input = transpile(
                    run_input, target_alias, **conversion_scheme_fields
                )

                if not (
                    isinstance(transpiled_run_input, list)
                    and all(
                        isinstance(item, (target_type, type(target_type)))
                        for item in cast(list, transpiled_run_input)
                    )
                ) and not isinstance(transpiled_run_input, (target_type, type(target_type))):
                    raise ProgramConversionError(
                        f"Expected transpile step to produce program of type of {target_type}, "
                        f"but instead got program of type {type(transpiled_run_input)}."
                    )

                return transpiled_run_input
            except (ProgramConversionError, ConversionPathNotFoundError) as err:
                cached_errors.append(err)

        if len(cached_errors) == 1:
            raise cached_errors[0]

        error_messages = "\n".join([str(error) for error in cached_errors])
        raise ProgramConversionError(
            "Transpile step failed after multiple attempts. "
            f"The following errors occurred:\n{error_messages}"
        )

    def transform(self, run_input: qbraid.programs.QPROGRAM) -> qbraid.programs.QPROGRAM:
        """
        Override this method with any runtime transformations to apply to the run
        input, e.g. circuit optimizations, device-specific gate set conversions, etc.
        Program input type should match output type.

        """
        return run_input

    def validate(
        self, run_input_batch: list[qbraid.programs.QPROGRAM], suppress_device_warning: bool = False
    ) -> None:
        """Verifies run input compatibility with target device.

        Raises:
            ProgramValidationError: If the run input is incompatible with the target device.
        """
        level = ValidationLevel(self._options.get("validate", 0))

        logger.debug("Validating run input for device at %s", level)

        if level == ValidationLevel.NONE:
            return None

        if not suppress_device_warning and self.status() != DeviceStatus.ONLINE:
            warnings.warn(
                "Device is not online. Submitting this job may result in an exception "
                "or a long wait time.",
                UserWarning,
            )

        for run_input in run_input_batch:
            try:
                program = load_program(run_input)
            except ProgramLoaderError:
                logger.info(
                    "Skipping qubit count validation: program type '%s' not supported natively.",
                    type(run_input).__name__,
                )
            else:
                if self.num_qubits and program.num_qubits > self.num_qubits:
                    message = (
                        f"Number of qubits in the circuit ({program.num_qubits}) exceeds "
                        f"the device's capacity ({self.num_qubits})."
                    )
                    if level == ValidationLevel.RAISE:
                        raise ProgramValidationError(message)
                    if level == ValidationLevel.WARN:
                        warnings.warn(message, UserWarning)

            if self._target_spec is None:
                continue

            target_spec = self._get_target_spec(run_input)

            try:
                target_spec.validate(run_input)
            except ValueError as err:
                if level == ValidationLevel.RAISE:
                    raise ProgramValidationError(err) from err
                if level == ValidationLevel.WARN:
                    warnings.warn(str(err), UserWarning)

        return None

    def prepare(self, run_input: qbraid.programs.QPROGRAM) -> Any:
        """Convert the quantum program to an intermediate representation (IR) compatible
        with the submission format required for the target device and its provider API."""
        if self._target_spec is None or not self._options.get("prepare"):
            return run_input

        logger.debug("Preparing (i.e. serializing) run input")

        target_spec = self._get_target_spec(run_input)
        return target_spec.serialize(run_input)

    def apply_runtime_profile(
        self, run_input: qbraid.programs.QPROGRAM
    ) -> qbraid.programs.QPROGRAM:
        """Process quantum program before passing to device run method.

        Returns:
            Transpiled and transformed quantum program
        """
        if self._target_spec is not None and self._options.get("transpile") is True:
            run_input_alias = get_program_type_alias(run_input, safe=True)
            run_input_spec = ProgramSpec(type(run_input), alias=run_input_alias)
            run_input = self.transpile(run_input, run_input_spec)

        is_single_output = not isinstance(run_input, list)
        run_input = [run_input] if is_single_output else run_input

        if self._options.get("transform") is True:
            logger.debug("Applying device-specific transformations (no-op in base class)")
            run_input = [self.transform(p) for p in cast(list, run_input)]

        self.validate(run_input)

        run_input = [self.prepare(p) for p in cast(list, run_input)]

        run_input = run_input[0] if is_single_output else run_input
        return run_input

    @abstractmethod
    def submit(
        self,
        run_input: Union[qbraid.programs.QPROGRAM, list[qbraid.programs.QPROGRAM]],
        *args,
        **kwargs,
    ) -> Union[qbraid.runtime.QuantumJob, list[qbraid.runtime.QuantumJob]]:
        """Vendor run method. Should return dictionary with the following keys."""

    def run(
        self,
        run_input: Union[qbraid.programs.QPROGRAM, list[qbraid.programs.QPROGRAM]],
        *args,
        **kwargs,
    ) -> Union[qbraid.runtime.QuantumJob, list[qbraid.runtime.QuantumJob]]:
        """
        Run a quantum job or a list of quantum jobs on this quantum device.

        Args:
            run_input: A single quantum program or a list of quantum programs to run on the device.

        Returns:
            A QuantumJob object or a list of QuantumJob objects corresponding to the input.
        """
        is_single_input = not isinstance(run_input, list)
        run_input = [run_input] if is_single_input else run_input
        run_input_compat = [self.apply_runtime_profile(program) for program in run_input]
        run_input_compat = run_input_compat[0] if is_single_input else run_input_compat
        logger.debug(
            "Submitting quantum program %s to device '%s'",
            "batch" if not is_single_input else "",
            self.id,
        )
        return self.submit(run_input_compat, *args, **kwargs)
